
__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class SSH(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)

        # Instruct the ExecWorkers to unset this environment variable.
        # Otherwise this will break nested SSH with SHELL spawner, i.e. when
        # both the sub-agent and CUs are started using SSH.
        self.env_removables.extend(["RP_SPAWNER_HOP"])


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        command = ru.which('ssh')

        if not command:
            raise RuntimeError("ssh not found!")

        # Some MPI environments (e.g. SGE) put a link to rsh as "ssh" into
        # the path.  We try to detect that and then use different arguments.
        if os.path.islink(command):

            target = os.path.realpath(command)

            if os.path.basename(target) == 'rsh':
                self._log.info('Detected that "ssh" is a link to "rsh".')
                return target

        command = '%s -o StrictHostKeyChecking=no -o ControlMaster=auto' % command

        self.launch_command = command


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_env     = cud.get('environment') or dict()
        task_args    = cud.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)

        if task_argstr: task_command = "%s %s" % (task_exec, task_argstr)
        else          : task_command = task_exec

        if not launch_script_hop :
            raise ValueError ("LMSSH.construct_command needs launch_script_hop!")

        if 'nodes' not in slots:
            raise RuntimeError('insufficient information to launch via %s: %s'
                              % (self.name, slots))

        if len(slots['nodes']) > 1:
            raise RuntimeError('ssh cannot run multinode units')

        host = slots['nodes'][0]['name']

        # Pass configured and available environment variables.
        # This is a crude version of env transplanting where we prep the
        # shell command line.  We likely won't survive any complicated vars
        # (multiline, quotes, etc)
        env_string  = ' '.join(['%s=%s' % (var, os.environ[var])
                                for var in self.EXPORT_ENV_VARIABLES
                                if  var in os.environ])
        env_string += ' '.join(['%s=%s' % (var, task_env[var])
                                for var in task_env])

        ssh_hop_cmd = "%s %s %s %s" % (self.launch_command, host, env_string,
                                       launch_script_hop)

        return task_command, ssh_hop_cmd


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"

import os

import radical.utils as ru

from ... import agent as rpa


# 'enum' for resource manager types
RM_NAME_FORK        = 'FORK'
RM_NAME_CCM         = 'CCM'
RM_NAME_LOADLEVELER = 'LOADLEVELER'
RM_NAME_LSF         = 'LSF'
RM_NAME_LSF_SUMMIT  = 'LSF_SUMMIT'
RM_NAME_PBSPRO      = 'PBSPRO'
RM_NAME_SGE         = 'SGE'
RM_NAME_SLURM       = 'SLURM'
RM_NAME_TORQUE      = 'TORQUE'
RM_NAME_COBALT      = 'COBALT'
RM_NAME_YARN        = 'YARN'
RM_NAME_SPARK       = 'SPARK'
RM_NAME_DEBUG       = 'DEBUG'


# ------------------------------------------------------------------------------
#
# Base class for ResourceManager implementations.
#
class ResourceManager(object):
    """
    The Resource Manager provides three fundamental information:

      ResourceManager.node_list      : list of node names
      ResourceManager.agent_node_list: list of nodes reserved for agent procs
      ResourceManager.cores_per_node : number of cores each node has available
      ResourceManager.gpus_per_node  : number of gpus  each node has available

    Schedulers can rely on these information to be available.  Specific
    ResourceManager incarnation may have additional information available -- but
    schedulers relying on those are invariably bound to the specific
    ResourceManager.  An example is the Torus Scheduler which relies on detailed
    torus layout information from the LoadLevelerRM (which describes the BG/Q).

    The ResourceManager will reserve nodes for the agent execution, by deriving
    the respectively required node count from the config's 'agents' section.
    Those nodes will be listed in ResourceManager.agent_node_list. Schedulers
    MUST NOT use the agent_node_list to place compute units -- CUs are limited
    to the nodes in ResourceManager.node_list.
    """

    # TODO: Core counts dont have to be the same number for all hosts.

    # TODO: We might not have reserved the whole node.

    # TODO: Given that the Agent can determine the real core count, in
    #       principle we could just ignore the config and use as many as we
    #       have to our availability (taken into account that we might not
    #       have the full node reserved of course)
    #       Answer: at least on Yellowstone this doesnt work for MPI,
    #               as you can't spawn more tasks then the number of slots.


    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self.name            = type(self).__name__
        self._cfg            = cfg
        self._session        = session
        self._log            = self._session._log
        self._prof           = self._session._prof
        self.requested_cores = self._cfg['cores']
        self.requested_gpus  = self._cfg['gpus']

        self._log.info("Configuring ResourceManager %s.", self.name)

        self.lm_info         = dict()
        self.rm_info         = dict()
        self.node_list       = list()
        self.agent_nodes     = dict()
        self.cores_per_node  = 0
        self.gpus_per_node   = 0
        self.lfs_per_node    = 0
        self.mem_per_node    = 0
        self.smt             = int(os.environ.get('RADICAL_SAGA_SMT', 1))

        # The ResourceManager will possibly need to reserve nodes for the agent,
        # according to the agent layout.  We dig out the respective requirements
        # from the config right here.
        self._agent_reqs = []
        agents = self._cfg.get('agents', {})

        # FIXME: this loop iterates over all agents *defined* in the layout, not
        #        over all agents which are to be actually executed, thus
        #        potentially reserving too many nodes.a
        # NOTE:  this code path is *within* the agent, so at least agent.0
        #        cannot possibly land on a different node.
        for agent in agents:
            target = agents[agent].get('target')
            # make sure that the target either 'local', which we will ignore,
            # or 'node'.
            if target == 'local':
                pass  # ignore that one
            elif target == 'node':
                self._agent_reqs.append(agent)
            else :
                raise ValueError("ill-formatted agent target '%s'" % target)

        # We are good to get rolling, and to detect the runtime environment of
        # the local ResourceManager.
        self._configure()
        self._log.info("Discovered execution environment: %s", self.node_list)

        # Make sure we got a valid nodelist and a valid setting for
        # cores_per_node
        if not self.node_list or self.cores_per_node < 1:
            raise RuntimeError('ResourceManager configuration invalid (%s)(%s)' %
                    (self.node_list, self.cores_per_node))

        # Check if the ResourceManager implementation reserved agent nodes.
        # If not, pick the first couple of nodes from the nodelist as fallback.
        if self._agent_reqs and not self.agent_nodes:
            self._log.info('Determine list of agent nodes generically.')
            for agent in self._agent_reqs:
                # Get a node from the end of the node list
                self.agent_nodes[agent] = self.node_list.pop()
                # If all nodes are taken by workers now, we can safely stop,
                # and let the raise below do its thing.
                if not self.node_list:
                    break

        if self.agent_nodes:
            self._log.info('Reserved nodes: %s' % list(self.agent_nodes.values()))
            self._log.info('Agent    nodes: %s' % list(self.agent_nodes.keys()))
            self._log.info('Worker   nodes: %s' % self.node_list)

        # Check if we can do any work
        if not self.node_list:
            raise RuntimeError('ResourceManager has no nodes left to run units')

        # After ResourceManager configuration, we call any existing config hooks
        # on the launch methods.  Those hooks may need to adjust the
        # ResourceManager settings (hello ORTE).  We only call LaunchMethod
        # hooks *once* (thus the set)
        launch_methods = set()
        launch_methods.add(self._cfg.get('mpi_launch_method'))
        launch_methods.add(self._cfg.get('task_launch_method'))
        launch_methods.add(self._cfg.get('agent_launch_method'))

        launch_methods.discard(None)

        for lm in launch_methods:
            try:
                ru.dict_merge(
                    self.lm_info,
                    rpa.LaunchMethod.rm_config_hook(name=lm,
                                                    cfg=self._cfg,
                                                    rm=self,
                                                    log=self._log,
                                                    profiler=self._prof))
            except Exception as e:
                # FIXME don't catch/raise
                self._log.exception(
                    "ResourceManager config hook failed: %s" % e)
                raise

            self._log.info("ResourceManager config hook succeeded (%s)" % lm)

        # For now assume that all nodes have equal amount of cores and gpus
        cores_avail = (len(self.node_list) + len(self.agent_nodes)) * self.cores_per_node
        gpus_avail  = (len(self.node_list) + len(self.agent_nodes)) * self.gpus_per_node

        # on debug runs, we allow more cpus/gpus to appear than physically exist
        if 'RADICAL_DEBUG' not in os.environ:
            if cores_avail < int(self.requested_cores):
                raise ValueError("Not enough cores available (%s < %s)."
                                % (str(cores_avail), str(self.requested_cores)))

            if gpus_avail < int(self.requested_gpus):
                raise ValueError("Not enough gpus available (%s < %s)."
                                % (str(gpus_avail), str(self.requested_gpus)))


        # NOTE: self.rm_info is what scheduler and launch method can
        # ultimately use, as it is included into the cfg passed to all
        # components.
        #
        # five elements are well defined:
        #   lm_info:        the dict received via the LM's rm_config_hook
        #   node_list:      a list of node names to be used for unit execution
        #   cores_per_node: as the name says
        #   gpus_per_node:  as the name says
        #   agent_nodes:    list of node names reserved for agent execution
        #
        # That list may turn out to be insufficient for some schedulers.  Yarn
        # for example may need to communicate YARN service endpoints etc.  an
        # ResourceManager can thus expand this dict, but is then likely bound to
        # a specific scheduler which can interpret the additional information.
        self.rm_info['name']           = self.name
        self.rm_info['lm_info']        = self.lm_info
        self.rm_info['node_list']      = self.node_list
        self.rm_info['cores_per_node'] = self.cores_per_node
        self.rm_info['gpus_per_node']  = self.gpus_per_node
        self.rm_info['agent_nodes']    = self.agent_nodes
        self.rm_info['lfs_per_node']   = self.lfs_per_node
        self.rm_info['mem_per_node']   = self.mem_per_node


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the
    # ResourceManager.
    #
    @classmethod
    def create(cls, name, cfg, session):

        from .ccm         import CCM
        from .fork        import Fork
        from .loadleveler import LoadLeveler
        from .lsf         import LSF
        from .lsf_summit  import LSF_SUMMIT
        from .pbspro      import PBSPro
        from .sge         import SGE
        from .slurm       import Slurm
        from .torque      import Torque
        from .cobalt      import Cobalt
        from .yarn        import Yarn
        from .spark       import Spark
        from .debug       import Debug

        # Make sure that we are the base-class!
        if cls != ResourceManager:
            raise TypeError("ResourceManager Factory only available to base class!")

        try:
            impl = {
                RM_NAME_FORK        : Fork,
                RM_NAME_CCM         : CCM,
                RM_NAME_LOADLEVELER : LoadLeveler,
                RM_NAME_LSF         : LSF,
                RM_NAME_LSF_SUMMIT  : LSF_SUMMIT,
                RM_NAME_PBSPRO      : PBSPro,
                RM_NAME_SGE         : SGE,
                RM_NAME_SLURM       : Slurm,
                RM_NAME_TORQUE      : Torque,
                RM_NAME_COBALT      : Cobalt,
                RM_NAME_YARN        : Yarn,
                RM_NAME_SPARK       : Spark,
                RM_NAME_DEBUG       : Debug
            }[name]
            return impl(cfg, session)

        except KeyError as e:
            raise RuntimeError("ResourceManager '%s' unknown" % name) from e


    # --------------------------------------------------------------------------
    #
    def stop(self):

        # During ResourceManager termination, we call any existing shutdown hooks on the
        # launch methods.  We only call LaunchMethod shutdown hooks *once*
        launch_methods = set()
        launch_methods.add(self._cfg.get('mpi_launch_method'))
        launch_methods.add(self._cfg.get('task_launch_method'))
        launch_methods.add(self._cfg.get('agent_launch_method'))

        launch_methods.discard(None)

        for lm in launch_methods:
            try:
                ru.dict_merge(
                    self.lm_info,
                    rpa.LaunchMethod.rm_shutdown_hook(name=lm,
                                                      cfg=self._cfg,
                                                      rm=self,
                                                      lm_info=self.lm_info,
                                                      log=self._log,
                                                      profiler=self._prof))
            except Exception as e:
                self._log.exception(
                    "ResourceManager shutdown hook failed: %s" % e)
                raise

            self._log.info("ResourceManager shutdown hook succeeded (%s)" % lm)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        raise NotImplementedError("_Configure missing for %s" % self.name)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import subprocess

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
# The Launch Method Implementation for Running YARN applications
#
class Yarn(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_config_hook(cls, name, cfg, rm, log, profiler=None):
        """
        this config hook will inspect the ResourceManager nodelist and, if
        needed, will start the YARN cluster on node[0].
        """
        log.info('Hook called by YARN ResourceManager with the name %s' % rm.name)

        def config_core_site(node):

            core_site_file = open(os.getcwd() + '/hadoop/etc/hadoop/core-site.xml','r')
            lines = core_site_file.readlines()
            core_site_file.close()

            prop_str  = '<property>\n'
            prop_str += '  <name>fs.default.name</name>\n'
            prop_str += '    <value>hdfs://%s:54170</value>\n' % node
            prop_str += '</property>\n'

            lines.insert(-1, prop_str)

            core_site_file = open(os.getcwd() + '/hadoop/etc/hadoop/core-site.xml', 'w')
            for line in lines:
                core_site_file.write(line)
            core_site_file.close()

        def config_hdfs_site():

            hdfs_site_file = open(os.getcwd() + '/hadoop/etc/hadoop/hdfs-site.xml', 'r')
            lines = hdfs_site_file.readlines()
            hdfs_site_file.close()

            prop_str  = '<property>\n'
            prop_str += ' <name>dfs.replication</name>\n'
            prop_str += ' <value>1</value>\n'
            prop_str += '</property>\n'

            prop_str += '<property>\n'
            prop_str += '  <name>dfs.name.dir</name>\n'
            prop_str += '    <value>file:///tmp/hadoop/hadoopdata/hdfs/namenode</value>\n'
            prop_str += '</property>\n'

            prop_str += '<property>\n'
            prop_str += '  <name>dfs.data.dir</name>\n'
            prop_str += '    <value>file:///tmp/hadoop/hadoopdata/hdfs/datanode</value>\n'
            prop_str += '</property>\n'

            lines.insert(-1, prop_str)

            hdfs_site_file = open(os.getcwd() + '/hadoop/etc/hadoop/hdfs-site.xml', 'w')
            for line in lines:
                hdfs_site_file.write(line)
            hdfs_site_file.close()

        def config_mapred_site():

            mapred_site_file = open(os.getcwd() + '/hadoop/etc/hadoop/mapred-site.xml.template', 'r')
            lines = mapred_site_file.readlines()
            mapred_site_file.close()

            prop_str  = ' <property>\n'
            prop_str += '  <name>mapreduce.framework.name</name>\n'
            prop_str += '   <value>yarn</value>\n'
            prop_str += ' </property>\n'

            lines.insert(-1,prop_str)

            mapred_site_file = open(os.getcwd() + '/hadoop/etc/hadoop/mapred-site.xml', 'w')
            for line in lines:
                mapred_site_file.write(line)
            mapred_site_file.close()

        def config_yarn_site(cores,nodelist,hostname):

            yarn_site_file = open(os.getcwd() + '/hadoop/etc/hadoop/yarn-site.xml', 'r')
            lines = yarn_site_file.readlines()
            yarn_site_file.close()

            total_mem_str = subprocess.check_output(['grep', 'MemTotal', '/proc/meminfo'])
            total_free_mem = int(total_mem_str.split()[1]) / 1048

            if nodelist.__len__() == 1:
                cores_used = cores / 2
                total_mem = total_free_mem * 0.75
            else:
                cores_used = cores * (len(nodelist) - 1)
                total_mem = total_free_mem * (len(nodelist) - 1)
                slaves = open(os.getcwd() + '/hadoop/etc/hadoop/slaves', 'w')
                for node in nodelist[1:]:
                    slaves.write('%s\n' % (node + hostname))
                slaves.close()
                master = open(os.getcwd() + '/hadoop/etc/hadoop/masters', 'w')
                master.write('%s\n' % (nodelist[0] + hostname))
                master.close()

            max_app_mem = total_mem / cores_used

            prop_str  = ' <property>\n'
            prop_str += '  <name>yarn.nodemanager.aux-services</name>\n'
            prop_str += '    <value>mapreduce_shuffle</value>\n'
            prop_str += ' </property>\n'

            prop_str += ' <property>\n'
            prop_str += '  <name>yarn.scheduler.maximum-allocation-mb</name>\n'
            prop_str += '   <value>%d</value>\n' % max_app_mem
            prop_str += ' </property>\n'

            prop_str += ' <property>\n'
            prop_str += '  <name>yarn.resourcemanager.hostname</name>\n'
            prop_str += '   <value>%s</value>\n' % (nodelist[0] + hostname)
            prop_str += ' </property>\n'

            prop_str += ' <property>\n'
            prop_str += '  <name>yarn.nodemanager.resource.cpu-vcores</name>\n'
            prop_str += '   <value>%d</value>\n' % cores_used
            prop_str += ' </property>\n'

            prop_str += ' <property>\n'
            prop_str += '  <name>yarn.nodemanager.resource.memory-mb</name>\n'
            prop_str += '   <value>%d</value>\n' % total_mem
            prop_str += ' </property>\n'

            lines.insert(-1, prop_str)

            yarn_site_file = open(os.getcwd() + '/hadoop/etc/hadoop/yarn-site.xml', 'w')
            for line in lines:
                yarn_site_file.write(line)
            yarn_site_file.close()

            scheduler_file = open(os.getcwd() + '/hadoop/etc/hadoop/capacity-scheduler.xml', 'r')
            lines = scheduler_file.readlines()
            scheduler_file.close()

            for line in lines:
                if line.startswith('    <value>org.apache.hadoop.yarn.util.resource.'):
                    new_line = '    <value>org.apache.hadoop.yarn.util.resource.' + 'DefaultResourceCalculator</value>\n'
                    lines[lines.index(line)] = new_line
                elif line.startswith('    <name>yarn.scheduler.capacity.maximum-am-resource-percent</name>'):
                    new_line = '    <value>1</value>\n'
                    lines[lines.index(line) + 1] = new_line

            scheduler_file = open(os.getcwd() + '/hadoop/etc/hadoop/capacity-scheduler.xml', 'w')
            for line in lines:
                scheduler_file.write(line)

            scheduler_file.close()

        # If the ResourceManager used is not YARN the namenode url is going to
        # be the first node in the list and the port is the default one, else
        # it is the one that the YARN ResourceManager returns
        hadoop_home = None
        if rm.name == 'YARNRM':  # FIXME: use constant
            log.info('Hook called by YARN ResourceManager')
            log.info('NameNode: %s', rm.namenode_url)
            service_url    = rm.namenode_url
            rm_url         = "%s:%s" % (rm.rm_ip, rm.rm_port)
            rm_ip          = rm.rm_ip
            launch_command = ru.which('yarn')

        else:
            # Here are the necessary commands to start the cluster.
            if rm.node_list[0] == 'localhost':
                node_name = rm.node_list[0]
            else:
                node = subprocess.check_output('/bin/hostname')
                log.info('Entered Else creation')
                node_name = node.split('\n')[0]

            # Download the tar file
            stat = os.system("wget http://apache.claz.org/hadoop/common/hadoop-2.6.5/hadoop-2.6.5.tar.gz")
            stat = os.system('tar xzf hadoop-2.6.5.tar.gz;mv hadoop-2.6.5 hadoop;rm -rf hadoop-2.6.5.tar.gz')
            # TODO: Decide how the agent will get Hadoop tar ball.
            log.debug('YARN Download Completed with code: %s', stat)


            hadoop_home        = os.getcwd() + '/hadoop'
            yarn_home          = hadoop_home

            # ------------------------------------------------------------------
            # Solution to find Java's home folder:
            # http://stackoverflow.com/questions/1117398/java-home-directory

            java = ru.which('java')
            if java != '/usr/bin/java':
                jpos = java.split('bin')
            else:
                jpos = os.path.realpath('/usr/bin/java').split('bin')

            if jpos[0].find('jre') != -1:
                java_home = jpos[0][:jpos[0].find('jre')]
            else:
                java_home = jpos[0]

            hadoop_env_file = open(hadoop_home + '/etc/hadoop/hadoop-env.sh', 'r')
            hadoop_env_file_lines = hadoop_env_file.readlines()
            hadoop_env_file.close()
            hadoop_env_file_lines[24] = 'export JAVA_HOME=%s' % java_home
            hadoop_env_file = open(hadoop_home + '/etc/hadoop/hadoop-env.sh', 'w')
            for line in hadoop_env_file_lines:
                hadoop_env_file.write(line)
            hadoop_env_file.close()
            host = node_name.split(rm.node_list[0])[1]

            config_core_site(node_name)
            config_hdfs_site()
            config_mapred_site()
            config_yarn_site(rm.cores_per_node, rm.node_list, host)  # FIXME GPU

            log.info('Start Formatting DFS')
            namenode_format = os.system(hadoop_home + '/bin/hdfs namenode -format -force')
            log.info('DFS Formatted. Starting DFS: %s', namenode_format)
            log.info('Starting YARN')
            yarn_start = subprocess.check_output([hadoop_home + '/sbin/start-all.sh'])
            if 'Error' in yarn_start:
                raise RuntimeError('Unable to start YARN cluster: %s'
                    % (yarn_start))
            else:
                log.info('Started YARN')

            # ------------------------------------------------------------------
            # Creating user's HDFS home folder
            log.debug('Running: %s/bin/hdfs dfs -mkdir /user' % hadoop_home)
            os.system('%s/bin/hdfs dfs -mkdir /user' % hadoop_home)
            uname = subprocess.check_output('whoami').split('\n')[0]
            log.debug('Running: %s/bin/hdfs dfs -mkdir /user/%s' % (hadoop_home,uname))
            os.system('%s/bin/hdfs dfs -mkdir /user/%s' % (hadoop_home,uname))
            check = subprocess.check_output(['%s/bin/hdfs' % hadoop_home,'dfs', '-ls', '/user'])
            log.info(check)
            log.info('Getting YARN app')
            os.system('wget https://www.dropbox.com/s/9yxbj9btibgtg40/Pilot-YARN-0.1-jar-with-dependencies.jar')

            # FIXME YARN: why was the scheduler configure called here?  Configure
            #             is already called during scheduler instantiation
            # self._scheduler._configure()

            service_url = node_name + ':54170'
            rm_url      = node_name
            launch_command = yarn_home + '/bin/yarn'
            rm_ip = node_name


        # The ResourceManager instance is only available here -- everything
        # which is later needed by the scheduler or launch method is stored in
        # an 'lm_info' dict.  That lm_info dict will be attached to the
        # scheduler's rm_info dict, and will be passed around as part of the
        # slots structure, so it is available on all LaunchMethod create_command
        # calls.
        lm_info = {'service_url'   : service_url,
                   'rm_url'        : rm_url,
                   'hadoop_home'   : hadoop_home,
                   'rm_ip'         : rm_ip,
                   'name'          : rm.name,
                   'launch_command': launch_command,
                   'nodename'      : rm.node_list[0] }

        return lm_info


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_shutdown_hook(cls, name, cfg, rm, lm_info, log, profiler=None):
        if 'name' not in lm_info:
            raise RuntimeError('name not in lm_info for %s' % name)

        if lm_info['name'] != 'YARNRM':  # FIXME: use constant
            log.info('Stoping YARN')
            os.system(lm_info['hadoop_home'] + '/sbin/stop-yarn.sh')

            log.info('Stoping DFS.')
            os.system(lm_info['hadoop_home'] + '/sbin/stop-dfs.sh')

            log.info("Deleting HADOOP files from temp")
            os.system('rm -rf /tmp/hadoop*')
            os.system('rm -rf /tmp/Jetty*')
            os.system('rm -rf /tmp/hsperf*')


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # Single Node configuration
        # FIXME : Upload App to another server, which will be always alive
        self._log.info(self._cfg['rm_info']['lm_info'])
        self.launch_command = self._cfg['rm_info']['lm_info']['launch_command']
        self._log.info('YARN was called')


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        work_dir     = cu['unit_sandbox_path']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_cores   = cud['cpu_processes'] * cud['cpu_threads']
        task_env     = cud.get('environment') or {}
        task_args    = cud.get('arguments')   or []
        task_argstr  = self._create_arg_string(task_args)

        # Construct the args_string which is the arguments given as input to the
        # shell script. Needs to be a string
        self._log.debug("Constructing YARN command")
        self._log.debug('Slots : %s', slots)

        if 'lm_info' not in slots:
            raise RuntimeError('No lm_info to launch via %s: %s'
                    % (self.name, slots))

        if not slots['lm_info']:
            raise RuntimeError('lm_info missing for %s: %s'
                               % (self.name, slots))

        if 'service_url' not in slots['lm_info']:
            raise RuntimeError('service_url not in lm_info for %s: %s'
                    % (self.name, slots))

        if 'rm_url' not in slots['lm_info']:
            raise RuntimeError('rm_url not in lm_info for %s: %s'
                    % (self.name, slots))


        if 'nodename' not in slots['lm_info']:
            raise RuntimeError('nodename not in lm_info for %s: %s'
                    % (self.name, slots))

        service_url = slots['lm_info']['service_url']
        rm_url      = slots['lm_info']['rm_url']
        client_node = slots['lm_info']['nodename']

        # ----------------------------------------------------------------------
        # Create YARN script
        # This funcion creates the necessary script for the execution of the
        # CU's workload in a YARN application. The function is responsible
        # to set all the necessary variables, stage in, stage out and create
        # the execution command that will run in the distributed shell that
        # the YARN application provides. There reason for staging out is
        # because after the YARN application has finished everything will be
        # deleted.

        print_str  = "echo '#!/usr/bin/env bash'>>ExecScript.sh\n"
        print_str += "echo ''>>ExecScript.sh\n"
        print_str += "echo ''>>ExecScript.sh\n"
        print_str += "echo '#---------------------------------------------------------'>>ExecScript.sh\n"
        print_str += "echo '# Staging Input Files'>>ExecScript.sh\n"

        self._log.debug('Creating input staging')
        if cud.get('input_staging'):
            scp_input_files = '"'
            for InputFile in cud['input_staging']:
                scp_input_files += '%s/%s ' % (work_dir, InputFile['target'])
            scp_input_files += '"'
            print_str += "echo 'start=""`date +%s.%3N`""'>>ExecScript.sh\n"
            print_str += "echo 'scp $YarnUser@%s:%s .'>>ExecScript.sh\n" % (client_node,scp_input_files)
            print_str += "echo 'stop=""`date +%s.%3N`""'>>ExecScript.sh\n"
            print_str += "echo 'time_spent=""$(echo ""$stop - $start"" | bc)""'>>ExecScript.sh\n"
            print_str += "echo 'echo $time_spent >>Yprof'>>ExecScript.sh\n"

        if cud.get('pre_exec'):
            pre_exec_string = ''
            for elem in cud['pre_exec']:
                pre_exec_string += '%s;' % elem
            pre_exec_string += ''
            print_str += "echo ''>>ExecScript.sh\n"
            print_str += "echo ''>>ExecScript.sh\n"
            print_str += "echo '#---------------------------------------------------------'>>ExecScript.sh\n"
            print_str += "echo '# Pre exec'>>ExecScript.sh\n"
            print_str += "echo '%s'>>ExecScript.sh\n" % pre_exec_string

        print_str += "echo ''>>ExecScript.sh\n"
        print_str += "echo ''>>ExecScript.sh\n"
        print_str += "echo '#---------------------------------------------------------'>>ExecScript.sh\n"
        print_str += "echo '# Creating Executing Command'>>ExecScript.sh\n"
        print_str += "echo 'start=""`date +%s.%3N`""'>>ExecScript.sh\n"
        print_str += "echo '%s %s 1>Ystdout 2>Ystderr'>>ExecScript.sh\n" % (task_exec, task_argstr)
        print_str += "echo 'stop=""`date +%s.%3N`""'>>ExecScript.sh\n"
        print_str += "echo 'time_spent=""$(echo ""$stop - $start"" | bc)""'>>ExecScript.sh\n"
        print_str += "echo 'echo $time_spent >>Yprof'>>ExecScript.sh\n"

        print_str += "echo ''>>ExecScript.sh\n"
        print_str += "echo ''>>ExecScript.sh\n"
        print_str += "echo '#---------------------------------------------------------'>>ExecScript.sh\n"
        print_str += "echo '# Staging Output Files'>>ExecScript.sh\n"
        print_str += "echo 'start=""`date +%s.%3N`""'>>ExecScript.sh\n"
        print_str += "echo 'YarnUser=$(whoami)'>>ExecScript.sh\n"
        scp_output_files = 'Ystderr Ystdout'

        if cud.get('output_staging'):
            for OutputFile in cud['output_staging']:
                scp_output_files += ' %s' % (OutputFile['source'])
        print_str += "echo 'scp -v %s $YarnUser@%s:%s'>>ExecScript.sh\n" % (scp_output_files, client_node, work_dir)
        print_str += "echo 'stop=""`date +%s.%3N`""'>>ExecScript.sh\n"
        print_str += "echo 'time_spent=""$(echo ""$stop - $start"" | bc)""'>>ExecScript.sh\n"
        print_str += "echo 'echo $time_spent >>Yprof'>>ExecScript.sh\n"
        print_str += "echo 'scp -v Yprof $YarnUser@%s:%s'>>ExecScript.sh\n" % (client_node, work_dir)

        print_str += "echo ''>>ExecScript.sh\n"
        print_str += "echo ''>>ExecScript.sh\n"
        print_str += "echo '#End of File'>>ExecScript.sh\n\n\n"

        env_string = ''
        for key,val in list(task_env.items()):
            env_string += '-shell_env ' + key + '=' + str(val) + ' '

        # app_name = '-appname '+ cud['uid']
        # Construct the ncores_string which is the number of cores used by the
        # container to run the script
        if task_cores:
            ncores_string = '-container_vcores ' + str(task_cores)
        else:
            ncores_string = ''

        # Construct the nmem_string which is the size of memory used by the
        # container to run the script
        # if task_nummem:
        #    nmem_string = '-container_memory '+task_nummem
        # else:
        #    nmem_string = ''

        # Getting the namenode's address.
        service_url = 'yarn://%s?fs=hdfs://%s' % (rm_url, service_url)

        yarn_command = '%s -jar ../Pilot-YARN-0.1-jar-with-dependencies.jar'\
                       ' com.radical.pilot.Client -jar ../Pilot-YARN-0.1-jar-with-dependencies.jar'\
                       ' -shell_script ExecScript.sh %s %s -service_url %s' % (self.launch_command,
                        env_string, ncores_string,service_url)

        self._log.debug("Yarn Command %s" % yarn_command)

        return print_str + yarn_command, None


# ------------------------------------------------------------------------------



#!/usr/bin/env python3
# pylint: disable=redefined-outer-name

import os
import sys

import radical.utils as ru
import radical.pilot as rp


# This script has to run as a task within an pilot allocation, and is
# a demonstration of a task overlay within the RCT framework.
# It will:
#
#   - create a master which bootstraps a specific communication layer
#   - insert n workers into the pilot (again as a task)
#   - perform RPC handshake with those workers
#   - send RPC requests to the workers
#   - terminate the worker
#
# The worker itself is an external program which is not covered in this code.


# ------------------------------------------------------------------------------
#
class MyMaster(rp.task_overlay.Master):
    '''
    This class provides the communication setup for the task overlay: it will
    set up the request / response communication queues and provide the endpoint
    information to be forwarded to the workers.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        # initialize the task overlay base class.  That base class will ensure
        # proper communication channels to the pilot agent.
        rp.task_overlay.Master.__init__(self, cfg=cfg)


    # --------------------------------------------------------------------------
    #
    def create_work_items(self):

        self._prof.prof('create_start')

        world_size = self._cfg.n_masters
        rank       = self._cfg.rank

        # create an initial list of work items to be distributed to the workers.
        # Work items MUST be serializable dictionaries.
        idx   = rank
        total = int(eval(self._cfg.workload.total))
        while idx < total:

            uid  = 'request.%06d' % idx
            item = {'uid' :   uid,
                    'mode':  'call',
                    'data': {'method': 'hello',
                             'kwargs': {'count': idx,
                                        'uid'  : uid}}}
            self.request(item)
            idx += world_size

        self._prof.prof('create_stop')


    # --------------------------------------------------------------------------
    #
    def result_cb(self, requests):

        # result callbacks can return new work items
        new_requests = list()
        for r in requests:
            sys.stdout.write('result_cb %s: %s [%s]\n' % (r.uid, r.state, r.result))
            sys.stdout.flush()

          # count = r.work['data']['kwargs']['count']
          # if count < 10:
          #     new_requests.append({'mode': 'call',
          #                          'data': {'method': 'hello',
          #                                   'kwargs': {'count': count + 100}}})

        return new_requests


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # This master script runs as a task within a pilot allocation.  The purpose
    # of this master is to (a) spawn a set or workers within the same
    # allocation, (b) to distribute work items (`hello` function calls) to those
    # workers, and (c) to collect the responses again.
    cfg_fname    = str(sys.argv[1])
    cfg          = ru.Config(cfg=ru.read_json(cfg_fname))
    cfg.rank     = int(sys.argv[2])

    n_workers  = cfg.n_workers
    cpn        = cfg.cpn
    gpn        = cfg.gpn
    descr      = cfg.worker_descr
    worker     = os.path.basename(cfg.worker)
    pwd        = os.getcwd()

    # add data staging to worker: link input_dir, impress_dir, and oe_license
    descr['arguments']     = [os.path.basename(worker)]
  # descr['input_staging'] = [
  #                            {'source': '%s/%s' % (pwd, worker),
  #                             'target': worker,
  #                             'action': rp.COPY,
  #                             'flags' : rp.DEFAULT_FLAGS,
  #                             'uid'   : 'sd.0'},
  #                            {'source': '%s/%s' % (pwd, cfg_fname),
  #                             'target': cfg_fname,
  #                             'action': rp.COPY,
  #                             'flags' : rp.DEFAULT_FLAGS,
  #                             'uid'   : 'sd.1'},
  #                           ]

    # one node is used by master.  Alternatively (and probably better), we could
    # reduce one of the worker sizes by one core.  But it somewhat depends on
    # the worker type and application workload to judge if that makes sense, so
    # we leave it for now.

    # create a master class instance - this will establish communication to the
    # pilot agent
    master = MyMaster(cfg)

    # insert `n` worker tasks into the agent.  The agent will schedule (place)
    # those workers and execute them.  Insert one smaller worker (see above)
    # NOTE: this assumes a certain worker size / layout
    print('workers: %d' % n_workers)
    master.submit(descr=descr, count=n_workers, cores=cpn,     gpus=gpn)
  # master.submit(descr=descr, count=1,         cores=cpn - 1, gpus=gpn)

    # wait until `m` of those workers are up
    # This is optional, work requests can be submitted before and will wait in
    # a work queue.
  # master.wait(count=nworkers)

    master.run()

    # simply terminate
    # FIXME: clean up workers


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class DPlace(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        LaunchMethod.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # dplace: job launcher for SGI systems (e.g. on Blacklight)
        self.launch_command = ru.which('dplace')


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_cores   = cud['cpu_processes']  # FIXME: also use cpu_threads
        task_args    = cud.get('arguments') or []
        task_argstr  = self._create_arg_string(task_args)

        if 'task_offsets' not in slots :
            raise RuntimeError('insufficient information to launch via %s: %s'
                               % (self.name, slots))

        # FIXME: This is broken due to changes lot structure
        task_offsets   = slots['task_offsets']
        assert(len(task_offsets) == 1)
        dplace_offset  = task_offsets[0]

        task_command   = "%s %s" % (task_exec, task_argstr)
        dplace_command = "%s -c %d-%d %s" % (self.launch_command, dplace_offset,
                                             dplace_offset + task_cores - 1,
                                             task_command)
        return dplace_command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

import radical.utils as ru

from .base import ResourceManager


# ------------------------------------------------------------------------------
#
class Cobalt(ResourceManager):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        ResourceManager.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # we only support Cobalt on Theta right now, and since we know that
        # Theta is a Cray, we know that aprun is available.  Alas, aprun
        # provides the only way (we could find so far) to determing the list of
        # nodes we have available (`COBALT_NODELIST` seems broken).  So we run
        # `aprun` with the rank of nodes we *think* we have, and with `-N 1` to
        # place one rank per node, and run `hostname` - that gives is the list
        # of hostnames.  The number of nodes we receive from `$COBALT_PARTSIZE`.

        n_nodes          = int(os.environ['COBALT_PARTSIZE'])
        out, _, _        = ru.sh_callout('aprun -n %d -N 1 hostname' % n_nodes)
        node_list        = out.split()
        assert(len(node_list) == n_nodes), node_list

        # we also want    to learn the core count per node
        cmd              = 'cat /proc/cpuinfo | grep processor | wc -l'
        out, _, _        = ru.sh_callout('aprun -n %d -N 1 %s' % (n_nodes, cmd))
        core_counts      = list(set([int(x) for x in out.split()]))
        assert(len(core_counts) == 1), core_counts
        cores_per_node   = core_counts[0]

        gpus_per_node    = self._cfg.get('gpus_per_node', 0)
        lfs_per_node     = {'path': ru.expand_env(
                                       self._cfg.get('lfs_path_per_node')),
                            'size':    self._cfg.get('lfs_size_per_node', 0)
                           }
        mem_per_node     = self._cfg.get('mem_per_node', 0)

        self._log.info("Found unique core counts: %s", cores_per_node)

        # node names are unique, so can serve as node uids
        self.node_list        = [[node, node] for node in node_list]
        self.cores_per_node   = cores_per_node
        self.gpus_per_node    = gpus_per_node
        self.lfs_per_node     = lfs_per_node
        self.mem_per_node     = mem_per_node


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import os
import sys

import radical.utils as ru
import radical.pilot as rp


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    cfg_file  = sys.argv[1]
    cfg_dir   = os.path.abspath(os.path.dirname(cfg_file))
    cfg_fname =                 os.path.basename(cfg_file)

    cfg       = ru.Config(cfg=ru.read_json(cfg_file))
    cpn       = cfg.cpn
    gpn       = cfg.gpn
    n_masters = cfg.n_masters
    n_workers = cfg.n_workers
    workload  = cfg.workload

    # each master uses a node, and each worker on each master uses a node
    nodes     =  n_masters + (n_masters * n_workers)
    print('nodes', nodes)

    master    = '%s/%s' % (cfg_dir, cfg.master)
    worker    = '%s/%s' % (cfg_dir, cfg.worker)

    session   = rp.Session()
    try:
        pd = rp.ComputePilotDescription(cfg.pilot_descr)
        pd.cores   = nodes * cpn
        pd.gpus    = nodes * gpn
        pd.runtime = cfg.runtime

        tds = list()

        for i in range(n_masters):
            td = rp.ComputeUnitDescription(cfg.master_descr)
            td.executable     = "python3"
            td.cpu_threads    = cpn
            td.gpu_processes  = gpn
            td.arguments      = [os.path.basename(master), cfg_file, i]
            td.input_staging  = [{'source': master,
                                  'target': os.path.basename(master),
                                  'action': rp.TRANSFER,
                                  'flags' : rp.DEFAULT_FLAGS},
                                 {'source': worker,
                                  'target': os.path.basename(worker),
                                  'action': rp.TRANSFER,
                                  'flags' : rp.DEFAULT_FLAGS},
                                 {'source': cfg_file,
                                  'target': os.path.basename(cfg_file),
                                  'action': rp.TRANSFER,
                                  'flags' : rp.DEFAULT_FLAGS}
                                ]
            tds.append(td)

        pmgr  = rp.PilotManager(session=session)
        umgr  = rp.UnitManager(session=session)
        pilot = pmgr.submit_pilots(pd)
        task  = umgr.submit_units(tds)

        umgr.add_pilots(pilot)
        umgr.wait_units()

    finally:
        session.close(download=True)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class JSRUN(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self.launch_command = ru.which('jsrun')
        assert(self.launch_command)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_config_hook(cls, name, cfg, rm, log, profiler):

        if 'session.lassen' in cfg['sid'].lower():
            # correctness of GPU IDs is based on env var CUDA_VISIBLE_DEVICES
            # which value is taken from `gpu_map`, it is set at class
            # `radical.pilot.agent.scheduler.base.AgentSchedulingComponent`
            # (method `_handle_cuda`)
            #
            # *) since the launching happens at the login node Lassen@LLNL
            #    thus the session name can be used to identify the machine
            #
            # FIXME: the `cvd_id_mode` setting should eventually move into the
            #        resource config.
            lm_info = {'cvd_id_mode': 'physical'}
        else:
            lm_info = {'cvd_id_mode': 'logical'}
        return lm_info


    # --------------------------------------------------------------------------
    #
    def _create_resource_set_file(self, slots, uid, sandbox):
        """
        This method takes as input a CU slots and creates the necessary
        resource set file. This resource set file is then used by jsrun to
        place and execute tasks on nodes.

        An example of a resource file is:

        * Task 1: 2 MPI procs, 2 threads per process and 2 gpus per process*

            rank 0 : {host: 1; cpu:  {0, 1}; gpu: {0,1}}
            rank 1 : {host: 1; cpu: {22,23}; gpu: {3,4}}

        * Task 2: 2 MPI procs, 1 thread per process and 1 gpus per process*

            rank 0 : {host: 2; cpu:  7; gpu: 2}
            rank 1 : {host: 2; cpu: 30; gpu: 5}

        * Task 3: 1 proc, 1 thread per process*

            1 : {host: 2; cpu:  7}

        Parameters
        ----------
        slots : List of dictionaries.

            The slots that the unit will be placed. A slot has the following
            format:

            {"nodes"         : [{"name"    : "a",
                                 "uid"     : 1,
                                 "gpu_map" : [],
                                 "core_map": [[0]],
                                 "lfs"     : {"path": "/dev/null", "size": 0}
                                }],
             "cores_per_node": 16,
             "gpus_per_node" : 6
             "lfs_per_node"  : {"size": 0, "path": "/dev/null"},
             "lm_info"       : "INFO",
            }

        uid     : unit ID (string)
        sandbox : unit sandbox (string)
        mpi     : MPI or not (bool, default: False)

        """

        # if `cpu_index_using: physical` is set to run at Lassen@LLNL,
        #  then it returns an error "error in ptssup_mkcltsock_afunix()"
        if slots['nodes'][0]['name'].lower().startswith('lassen'):
            rs_str = ''
        else:
            rs_str = 'cpu_index_using: physical\n'
        rank = 0
        for node in slots['nodes']:

            gpu_maps = list(node['gpu_map'])
            for map_set in node['core_map']:
                cores = ','.join(str(core) for core in map_set)
                rs_str += 'rank: %d: {'  % rank
                rs_str += ' host: %s;'  % str(node['uid'])
                rs_str += ' cpu: {%s}'  % cores
                if gpu_maps:
                    gpus = [str(gpu_map[0]) for gpu_map in gpu_maps]
                    rs_str += '; gpu: {%s}' % ','.join(gpus)
                rs_str += '}\n'
                rank   += 1

        rs_name = '%s/%s.rs' % (sandbox, uid)
        with open(rs_name, 'w') as fout:
            fout.write(rs_str)

        return rs_name


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        uid          = cu['uid']
        slots        = cu['slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_args    = cud.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)
        task_sandbox = cu['unit_sandbox_path']

        assert(slots), 'missing slots for %s' % uid

        self._log.debug('prep %s', uid)

        if task_argstr: task_command = "%s %s" % (task_exec, task_argstr)
        else          : task_command = task_exec

        env_string = ''
      # task_env   = cud.get('environment') or dict()
      # env_list   = self.EXPORT_ENV_VARIABLES + task_env.keys()
      # env_string = ' '.join(['-E "%s"' % var for var in env_list])
      #
      # # jsrun fails if an -E export is not set
      # for var in env_list:
      #     if var not in os.environ:
      #         os.environ[var] = ''

        # from https://www.olcf.ornl.gov/ \
        #             wp-content/uploads/2018/11/multi-gpu-workshop.pdf
        #
        # CUDA with    MPI, use jsrun --smpiargs="-gpu"
        # CUDA without MPI, use jsrun --smpiargs="off"
        #
        # We only set this for CUDA tasks
        if 'cuda' in cud.get('gpu_thread_type', '').lower():
            if 'mpi' in cud.get('gpu_process_type', '').lower():
                smpiargs = '--smpiargs="-gpu"'
            else:
                smpiargs = '--smpiargs="off"'
        else:
            smpiargs = ''

        rs_fname = self._create_resource_set_file(slots=slots, uid=uid,
                                                  sandbox=task_sandbox)

      # flags = '-n%d -a1 ' % (task_procs)
        command = '%s --erf_input %s %s %s %s' % (self.launch_command, rs_fname,
                                                  smpiargs, env_string,
                                                  task_command)

      # with open('./commands.log', 'a') as fout:
      #     fout.write('%s\n' % command)

        return command, None


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import radical.pilot as rp


verbose  = os.environ.get('RADICAL_PILOT_VERBOSE', 'REPORT')
os.environ['RADICAL_PILOT_VERBOSE'] = verbose

""" DESCRIPTION: Tutorial 2: Chaining Tasks.
For every task A_n a task B_n is started consecutively.
"""

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    RESOURCE_LABEL  = None
    PILOT_CORES     = None
    NUMBER_CHAINS   = None
    CU_A_EXECUTABLE = None
    CU_B_EXECUTABLE = None
    QUEUE           = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # Change the user name below if you are using a remote resource
        # and your username on that resource is different from the username
        # on your local machine.
        #


        # Add a Pilot Manager. Pilot managers manage one or more ComputePilots.
        print("Initializing Pilot Manager ...")
        pmgr = rp.PilotManager(session=session)


        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # If you want to run this example on your local machine, you don't have
        # to change anything here.
        #
        # Change the resource below if you want to run on a remote resource.
        # You also might have to set the 'project' to your allocation ID if
        # your remote resource does compute time accounting.
        #
        # A list of preconfigured resources can be found at:
        # https://radicalpilot.readthedocs.io/en/stable/ \
        #        machconf.html#preconfigured-resources
        #
        pdesc = rp.ComputePilotDescription ()
        pdesc.resource = RESOURCE_LABEL
        pdesc.runtime  = 30
        pdesc.cores    = PILOT_CORES
        pdesc.cleanup  = True

        # submit the pilot.
        print("Submitting Compute Pilot to Pilot Manager ...")
        pilot = pmgr.submit_pilots(pdesc)

        # Combine the ComputePilot, the ComputeUnits and a scheduler via
        # a UnitManager object.
        print("Initializing Unit Manager ...")
        umgr = rp.UnitManager (session=session)


        # Add the created ComputePilot to the UnitManager.
        print("Registering Compute Pilot with Unit Manager ...")
        umgr.add_pilots(pilot)

        # submit A cus to pilot job
        cudesc_list_A = []
        for i in range(NUMBER_CHAINS):

            # -------- BEGIN USER DEFINED CU A_n DESCRIPTION --------- #
            cudesc = rp.ComputeUnitDescription()
            cudesc.environment = {"CU_LIST": "A", "CU_NO": "%02d" % i}
            cudesc.executable  = CU_A_EXECUTABLE
            cudesc.arguments   = ['"$CU_LIST CU with id $CU_NO"']
            cudesc.cores       = 1
            # -------- END USER DEFINED CU A_n DESCRIPTION --------- #

            cudesc_list_A.append(cudesc)

        # Submit the previously created ComputeUnit descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning ComputeUnits to the ComputePilots.
        print("Submit 'A' Compute Units to Unit Manager ...")
        cu_list_A = umgr.submit_units(cudesc_list_A)

        # Chaining cus i.e submit a compute unit, when compute unit from A is
        # successfully executed.  A B CU reads the content of the output file of
        # an A CU and writes it into its own output file.
        cu_list_B = []

        # We create a copy of cu_list_A so that we can remove elements from it,
        # and still reference to the original index.
        cu_list_A_copy = cu_list_A[:]
        while cu_list_A:
            for cu_a in cu_list_A:
                idx = cu_list_A_copy.index(cu_a)

                cu_a.wait ()
                print("'A' Compute Unit '%s' done. Submitting 'B' CU ..." % idx)

                # -------- BEGIN USER DEFINED CU B_n DESCRIPTION --------- #
                cudesc = rp.ComputeUnitDescription()
                cudesc.environment = {'CU_LIST': 'B', 'CU_NO': "%02d" % idx}
                cudesc.executable  = CU_B_EXECUTABLE
                cudesc.arguments   = ['"$CU_LIST CU with id $CU_NO"']
                cudesc.cores       = 1
                # -------- END USER DEFINED CU B_n DESCRIPTION --------- #

                # Submit CU to Pilot Job
                cu_b = umgr.submit_units(cudesc)
                cu_list_B.append(cu_b)
                cu_list_A.remove(cu_a)

        print("Waiting for 'B' Compute Units to complete ...")
        for cu_b in cu_list_B :
            cu_b.wait ()
            print("'B' Compute Unit '%s' finished with output:" % (cu_b.uid))
            print(cu_b.stdout)

        print("All Compute Units completed successfully!")

    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


# ------------------------------------------------------------------------------






import radical.utils as ru


# ------------------------------------------------------------------------------
#
class ClusterConfig(ru.Config):

    _schema = {
            'description'    : str,
            'lrms'           : str,

            'cores_per_node' : int,
            'gpus_per_node'  : int,
            'mem_per_node'   : int,
            'lfs_per_node'   : int,
            'lfs_path'       : str,
            }

    _defaults = {
            'description'    : '',
            'lrms'           : 'FORK',

            'cores_per_node' : 1,
            'gpus_per_node'  : 0,
            'mem_per_node'   : 0,
            'lfs_per_node'   : 0,
            'lfs_path'       : '/tmp/',
            }


# ------------------------------------------------------------------------------
#
class AccessConfigOption(ru.Config):

    _schema = {
            'job_manager'    : str,
            'job_manager_hop': str,
            'filesystem'     : str,
            }

    _defaults = {
            'job_manager'    : 'fork://localhost/',
            'job_manager_hop': None,
            'filesystem'     : 'file://localhost/',
            }


# ------------------------------------------------------------------------------
#
class AccessConfig(ru.Config):

    _schema = {
            'default'        : str,
            'options'        : {str: AccessConfigOption}
            }

    _defaults = {
            'default'        : 'fork',
            'options'        : {'fork' : AccessConfigOption()}
            }


# ------------------------------------------------------------------------------
#
class SystemConfigOption(ru.Config):

    _schema = {
            'pre_bootstrap_0': [str],
            'pre_bootstrap_1': [str],
            'pre_bootstrap_2': [str],
            'export_to_cu'   : [str],
            'cu_pre_exec'    : [str],
            }

    _defaults = {
            'pre_bootstrap_0': None,
            'pre_bootstrap_1': None,
            'pre_bootstrap_2': None,
            'export_to_cu'   : None,
            'cu_pre_exec'    : None,
            }


# ------------------------------------------------------------------------------
#
class SystemConfig(ru.Config):

    _schema = {
            'valid_roots'    : [str],
            'sandbox_base'   :  str,
            'python_dist'    :  str,
            'virtenv'        :  str,
            'virtenv_mode'   :  str,
            'rp_version'     :  str,
            'pre_bootstrap_0': [str],
            'pre_bootstrap_1': [str],
            'pre_bootstrap_2': [str],
            'export_to_cu'   : [str],
            'cu_pre_exec'    : [str],
            'options'        : {str: SystemConfigOption}
            }


    _defaults = {
            'valid_roots'    : ['/'],
            'sandbox_base'   : '$HOME',
            'python_dist'    : 'default',
            'virtenv'        : None,
            'virtenv_mode'   : 'create',
            'rp_version'     : 'local',
            'pre_bootstrap_0': None,
            'pre_bootstrap_1': None,
            'pre_bootstrap_2': None,
            'export_to_cu'   : None,
            'cu_pre_exec'    : None,
            'options'        : None,
            }


class AgentConfig(ru.Config):

    _schema = {
            'layout'         : str,
            'config'         : str,
            }

    _defaults = {
            'layout'         : 'default',
            'config'         : 'default',
            }


# ------------------------------------------------------------------------------
#
class ResourceConfig(ru.Config):

    _schema = {
            'agent'          : AgentConfig,
            'cluster'        : ClusterConfig,
            'access'         : AccessConfig,
            'system'         : SystemConfig,
            }

    _defaults = {
            'agent'          : AgentConfig(),
            'cluster'        : ClusterConfig(),
            'access'         : AccessConfig(),
            'system'         : SystemConfig(),
            }


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys
import random

import radical.pilot as rp
import radical.utils as ru

pwd = os.path.abspath(os.path.dirname(__file__))


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # use the resource specified as argument, fall back to localhost
    if   len(sys.argv) > 1: resource = sys.argv[1]
    else                  : resource = 'local.localhost'

    session = rp.Session()

    try:
        # read the config used for resource details
        report.info('read config')
        config = ru.read_json('%s/../config.json' % pwd)
        report.ok('>>ok\n')

        report.header('submit pilots')

        pd_init = {'resource'      : resource,
                   'runtime'       : 60,  # pilot runtime (min)
                   'exit_on_error' : True,
                   'cores'         : 32
                  }
        pdesc = rp.ComputePilotDescription(pd_init)
        pmgr  = rp.PilotManager(session=session)
        pilot = pmgr.submit_pilots(pdesc)

        report.header('submit bags of tasks')

        umgr = rp.UnitManager(session=session)
        umgr.add_pilots(pilot)

        # run N bags of tasks, where each bag contains M tasks of different
        # sizes.  All tasks within the same bag are to get scheduled on the
        # same node (colocated)

        n_bags    = 2
        bag_size  = 3
        task_size = [5, 1, 4]

        assert(len(task_size) == bag_size)

        cuds = list()
        for b in range(n_bags):
            for tid,s in enumerate(task_size):
                cud = rp.ComputeUnitDescription()
                cud.executable       = '%s/colocated_task.sh' % pwd
                cud.arguments        = [b, bag_size, tid]
                cud.cpu_processes    = s
                cud.cpu_process_type = rp.MPI
                cud.tags             = {'colocate': {'bag' : b,
                                                     'size': bag_size}}
                cud.name             =  'b%03d-t%03d' % (b, tid)
                print(cud.name)
                cuds.append(cud)
                report.progress()

        random.shuffle(cuds)

        umgr.submit_units(cuds)

        report.header('gather results')
        umgr.wait_units()


    except Exception as e:
        # Something unexpected happened in the pilot code above
        report.error('caught Exception: %s\n' % e)
        ru.print_exception_trace()
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        ru.print_exception_trace()
        report.warn('exit requested with %s\n' % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.  This will kill all remaining pilots.
        report.header('finalize')
        session.close(download=True)

    report.header()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2017, http://radical.rutgers.edu"
__license__   = "MIT"


import json

import radical.utils        as ru

from ...   import states    as rps
from ...   import constants as rpc

from ..launch_method import LaunchMethod
from .base           import AgentSchedulingComponent


# ------------------------------------------------------------------------------
#
class Flux(AgentSchedulingComponent):
    '''
    Pass all scheduling and execution control to Flux, and leave it to the
    execution component to collect state updates.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self.nodes = None

        AgentSchedulingComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        import flux

        flux_url   = self._cfg['rm_info']['lm_info']['flux_env']['FLUX_URI']
        self._flux = flux.Flux(url=flux_url)

        # don't advance tasks via the component's `advance()`, but push them
        # toward the executor *without state change* - state changes are
        # performed in retrospect by the executor, based on the scheduling and
        # execution events collected from Flux.
        qname   = rpc.AGENT_EXECUTING_QUEUE
        fname   = '%s/%s.cfg' % (self._cfg.path, qname)
        cfg     = ru.read_json(fname)
        self._q = ru.zmq.Putter(qname, cfg['put'])

        # create job spec via the flux LM
        self._lm = LaunchMethod.create(name    = 'FLUX',
                                       cfg     = self._cfg,
                                       session = self._session)


    # --------------------------------------------------------------------------
    #
    def work(self, units):

        # overload the base class work method

        from flux import job as flux_job

        self.advance(units, rps.AGENT_SCHEDULING, publish=True, push=False)

        for unit in units:

          # # FIXME: transfer from executor
          # self._cu_environment = self._populate_cu_environment()

            jd  = json.dumps(ru.read_json('/home/merzky/projects/flux/spec.json'))
            jid = flux_job.submit(self._flux, jd)
            unit['flux_id'] = jid

            # publish without state changes - those are retroactively applied
            # based on flux event timestamps.
            # TODO: apply some bulking, submission is not really fast.
            #       But at the end performance is determined by flux now, so
            #       communication only affects timelyness of state updates.
            self._q.put(unit)


  # # --------------------------------------------------------------------------
  # #
  # def _populate_cu_environment(self):
  #
  #     import tempfile
  #
  #     self.gtod   = "%s/gtod" % self._pwd
  #     self.tmpdir = tempfile.gettempdir()
  #
  #     # if we need to transplant any original env into the CU, we dig the
  #     # respective keys from the dump made by bootstrap_0.sh
  #     self._env_cu_export = dict()
  #     if self._cfg.get('export_to_cu'):
  #         with open('env.orig', 'r') as f:
  #             for line in f.readlines():
  #                 if '=' in line:
  #                     k,v = line.split('=', 1)
  #                     key = k.strip()
  #                     val = v.strip()
  #                     if key in self._cfg['export_to_cu']:
  #                         self._env_cu_export[key] = val
  #
  #
  # # --------------------------------------------------------------------------
  # #
  # def _populate_cu_environment(self):
  #     """Derive the environment for the cu's from our own environment."""
  #
  #     # Get the environment of the agent
  #     new_env = copy.deepcopy(os.environ)
  #
  #     #
  #     # Mimic what virtualenv's "deactivate" would do
  #     #
  #     old_path = new_env.pop('_OLD_VIRTUAL_PATH', None)
  #     if old_path:
  #         new_env['PATH'] = old_path
  #
  #     old_ppath = new_env.pop('_OLD_VIRTUAL_PYTHONPATH', None)
  #     if old_ppath:
  #         new_env['PYTHONPATH'] = old_ppath
  #
  #     old_home = new_env.pop('_OLD_VIRTUAL_PYTHONHOME', None)
  #     if old_home:
  #         new_env['PYTHON_HOME'] = old_home
  #
  #     old_ps = new_env.pop('_OLD_VIRTUAL_PS1', None)
  #     if old_ps:
  #         new_env['PS1'] = old_ps
  #
  #     new_env.pop('VIRTUAL_ENV', None)
  #
  #     # Remove the configured set of environment variables from the
  #     # environment that we pass to Popen.
  #     for e in list(new_env.keys()):
  #         for r in self._lm.env_removables:
  #             if e.startswith(r):
  #                 new_env.pop(e, None)
  #
  #     return new_env


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import errno
import shutil
import tarfile

import radical.saga  as rs
import radical.utils as ru

from ...  import utils     as rpu
from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentStagingInputComponent

from ...staging_directives import complete_url


# ------------------------------------------------------------------------------
#
class Default(AgentStagingInputComponent):
    """
    This component performs all agent side input staging directives for compute
    units.  It gets units from the agent_staging_input_queue, in
    AGENT_STAGING_INPUT_PENDING state, will advance them to AGENT_STAGING_INPUT
    state while performing the staging, and then moves then to the
    AGENT_SCHEDULING_PENDING state, into the agent_scheduling_queue.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentStagingInputComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_STAGING_INPUT_PENDING,
                            rpc.AGENT_STAGING_INPUT_QUEUE, self.work)

        self.register_output(rps.AGENT_SCHEDULING_PENDING,
                             rpc.AGENT_SCHEDULING_QUEUE)


    # --------------------------------------------------------------------------
    #
    def _work(self, units):

        # we first filter out any units which don't need any input staging, and
        # advance them again as a bulk.  We work over the others one by one, and
        # advance them individually, to avoid stalling from slow staging ops.

        no_staging_units = list()
        staging_units    = list()

        for unit in units:

            # check if we have any staging directives to be enacted in this
            # component
            actionables = list()
            for sd in unit['description'].get('input_staging', []):

                if sd['action'] in [rpc.LINK, rpc.COPY, rpc.MOVE, rpc.TARBALL]:
                    actionables.append(sd)

            if actionables:
                staging_units.append([unit, actionables])
            else:
                no_staging_units.append(unit)

        if no_staging_units:
            self.advance(no_staging_units, rps.AGENT_SCHEDULING_PENDING,
                         publish=True, push=True)

        for unit,actionables in staging_units:
            self._handle_unit(unit, actionables)


    # --------------------------------------------------------------------------
    #
    def _handle_unit(self, unit, actionables):

        uid = unit['uid']

        # By definition, this compoentn lives on the pilot's target resource.
        # As such, we *know* that all staging ops which would refer to the
        # resource now refer to file://localhost, and thus translate the unit,
        # pilot and resource sandboxes into that scope.  Some assumptions are
        # made though:
        #
        #   * paths are directly translatable across schemas
        #   * resource level storage is in fact accessible via file://
        #
        # FIXME: this is costly and should be cached.

        unit_sandbox     = ru.Url(unit['unit_sandbox'])
        pilot_sandbox    = ru.Url(unit['pilot_sandbox'])
        resource_sandbox = ru.Url(unit['resource_sandbox'])

        unit_sandbox.schema     = 'file'
        pilot_sandbox.schema    = 'file'
        resource_sandbox.schema = 'file'

        unit_sandbox.host       = 'localhost'
        pilot_sandbox.host      = 'localhost'
        resource_sandbox.host   = 'localhost'

        src_context = {'pwd'      : str(unit_sandbox),       # !!!
                       'unit'     : str(unit_sandbox),
                       'pilot'    : str(pilot_sandbox),
                       'resource' : str(resource_sandbox)}
        tgt_context = {'pwd'      : str(unit_sandbox),       # !!!
                       'unit'     : str(unit_sandbox),
                       'pilot'    : str(pilot_sandbox),
                       'resource' : str(resource_sandbox)}


        # we can now handle the actionable staging directives
        for sd in actionables:

            action = sd['action']
            flags  = sd['flags']
            did    = sd['uid']
            src    = sd['source']
            tgt    = sd['target']

            self._prof.prof('staging_in_start', uid=uid, msg=did)

            assert(action in [rpc.COPY, rpc.LINK, rpc.MOVE,
                              rpc.TRANSFER, rpc.TARBALL])

            # we only handle staging which does *not* include 'client://' src or
            # tgt URLs - those are handled by the umgr staging components
            if src.startswith('client://') and action != rpc.TARBALL:
                self._log.debug('skip staging for src %s', src)
                self._prof.prof('staging_in_skip', uid=uid, msg=did)
                continue

            if tgt.startswith('client://'):
                self._log.debug('skip staging for tgt %s', tgt)
                self._prof.prof('staging_in_skip', uid=uid, msg=did)
                continue

            # Fix for when the target PATH is empty
            # we assume current directory is the unit staging 'unit://'
            # and we assume the file to be copied is the base filename
            # of the source
            if tgt is None: tgt = ''
            if tgt.strip() == '':
                tgt = 'unit:///{}'.format(os.path.basename(src))
            # Fix for when the target PATH is exists *and* it is a folder
            # we assume the 'current directory' is the target folder
            # and we assume the file to be copied is the base filename
            # of the source
            elif os.path.exists(tgt.strip()) and os.path.isdir(tgt.strip()):
                tgt = os.path.join(tgt, os.path.basename(src))


            src = complete_url(src, src_context, self._log)
            tgt = complete_url(tgt, tgt_context, self._log)

            # Currently, we use the same schema for files and folders.
            assert(tgt.schema == 'file'), 'staging tgt must be file://'

            if action in [rpc.COPY, rpc.LINK, rpc.MOVE]:
                assert(src.schema == 'file'), 'staging src expected as file://'

            # SAGA will take care of dir creation - but we do it manually
            # for local ops (copy, link, move)
            if flags & rpc.CREATE_PARENTS and action != rpc.TRANSFER:
                tgtdir = os.path.dirname(tgt.path)
                if tgtdir != unit_sandbox.path:
                    self._log.debug("mkdir %s", tgtdir)
                    rpu.rec_makedir(tgtdir)

            if action == rpc.COPY:
                try:
                    shutil.copytree(src.path, tgt.path)
                except OSError as exc:
                    if exc.errno == errno.ENOTDIR:
                        shutil.copy(src.path, tgt.path)
                    else:
                        raise

            elif action == rpc.LINK:

                # Fix issue/1513 if link source is file and target is folder.
                # should support POSIX standard where link is created
                # with the same name as the source
                if os.path.isfile(src.path) and os.path.isdir(tgt.path):
                    os.symlink(src.path,
                               '%s/%s' % (tgt.path, os.path.basename(src.path)))

                else:
                    os.symlink(src.path, tgt.path)

            elif action == rpc.MOVE:
                shutil.move(src.path, tgt.path)

            elif action == rpc.TRANSFER:

                # NOTE:  TRANSFER directives don't arrive here right now.
                # FIXME: we only handle srm staging right now, and only for
                #        a specific target proxy. Other TRANSFER directives are
                #        left to umgr input staging.  We should use SAGA to
                #        attempt all staging ops which do not involve the client
                #        machine.
                if src.schema == 'srm':
                    # FIXME: cache saga handles
                    srm_dir = rs.filesystem.Directory('srm://proxy/?SFN=bogus')
                    srm_dir.copy(src, tgt)
                    srm_dir.close()

                else:
                    self._log.error('no transfer for %s -> %s', src, tgt)
                    self._prof.prof('staging_in_fail', uid=uid, msg=did)
                    raise NotImplementedError('unsupported transfer %s' % src)

            elif action == rpc.TARBALL:

                # If somethig was staged via the tarball method, the tarball is
                # extracted and then removed from the unit folder.  The target
                # path is expected to be an *absolute* path on the target system
                # - any relative paths specified by the application are expected
                # to get expanded on the client side.
                tarball = '%s/%s.tar' % (os.path.dirname(tgt.path), uid)
                self._log.debug('extract tarball for %s', tarball)
                tar = tarfile.open(tarball)
                tar.extractall(path='/')
                tar.close()

              # FIXME: make tarball removal dependent on debug settings
              # os.remove(os.path.dirname(tgt.path) + '/' + uid + '.tar')

            self._prof.prof('staging_in_stop', uid=uid, msg=did)

        # all staging is done -- pass on to the scheduler
        self.advance(unit, rps.AGENT_SCHEDULING_PENDING, publish=True, push=True)


# ------------------------------------------------------------------------------





# pylint: disable=unused-argument    # W0613 Unused argument 'timeout' & 'input'
# pylint: disable=redefined-builtin  # W0622 Redefining built-in 'input'

import os
import copy
import time

import threading       as mt
import radical.utils   as ru

from ..          import constants      as rpc
from ..          import states         as rps


# ------------------------------------------------------------------------------
#
class ComponentManager(object):
    '''
    RP spans a hierarchy of component instances: the application has a pmgr and
    umgr, and the umgr has a staging component and a scheduling component, and
    the pmgr has a launching component, and components also can have bridges,
    etc. etc.  This ComponentManager centralises the code needed to spawn,
    manage and terminate such components - any code which needs to create
    component should create a ComponentManager instance and pass the required
    component and bridge layout and configuration.  Callng `stop()` on the cmgr
    will terminate the components and brisged.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        self._cfg  = ru.Config('radical.pilot.cmgr', cfg=cfg)
        self._sid  = self._cfg.sid
        self._uid  = ru.generate_id('cmgr', ns=self._sid)
        self._uids = [self._uid]  # uids to track hartbeats for (incl. own)

        self._prof = ru.Profiler(self._uid, ns='radical.pilot',
                               path=self._cfg.path)
        self._log  = ru.Logger(self._uid, ns='radical.pilot',
                               path=self._cfg.path)

        self._prof.prof('init2', uid=self._uid, msg=self._cfg.path)

        # Every ComponentManager runs a HB pubsub bridge in a separate thread.
        # That HB channel should be used by all components and bridges created
        # under this CMGR.
        bcfg = ru.Config(cfg={'channel'    : 'heartbeat',
                              'type'       : 'pubsub',
                              'uid'        : self._uid + '.hb',
                              'stall_hwm'  : 1,
                              'bulk_size'  : 0,
                              'path'       : self._cfg.path})
        self._hb_bridge = ru.zmq.PubSub(bcfg)
        self._hb_bridge.start()

        self._cfg.heartbeat.addr_pub = str(self._hb_bridge.addr_pub)
        self._cfg.heartbeat.addr_sub = str(self._hb_bridge.addr_sub)

        # runs a HB monitor on that channel
        self._hb = ru.Heartbeat(uid=self.uid,
                                timeout=self._cfg.heartbeat.timeout,
                                interval=self._cfg.heartbeat.interval,
                                beat_cb=self._hb_beat_cb,  # on every heartbeat
                                term_cb=self._hb_term_cb,  # on termination
                                log=self._log)

        self._hb_pub = ru.zmq.Publisher('heartbeat',
                                        self._cfg.heartbeat.addr_pub,
                                        log=self._log, prof=self._prof)
        self._hb_sub = ru.zmq.Subscriber('heartbeat',
                                         self._cfg.heartbeat.addr_sub,
                                         topic='heartbeat', cb=self._hb_sub_cb,
                                         log=self._log, prof=self._prof)

        # confirm the bridge being usable by listening to our own heartbeat
        self._hb.start()
        self._hb.wait_startup(self._uid, self._cfg.heartbeat.timeout)
        self._log.info('heartbeat system up')


    # --------------------------------------------------------------------------
    #
    def _hb_sub_cb(self, topic, msg):
        '''
        keep track of heartbeats for all bridges/components we know
        '''

      # self._log.debug('hb_sub %s: get %s check', self.uid, msg['uid'])
        if msg['uid'] in self._uids:
          # self._log.debug('hb_sub %s: get %s used', self.uid, msg['uid'])
            self._hb.beat(uid=msg['uid'])


    # --------------------------------------------------------------------------
    #
    def _hb_beat_cb(self):
        '''
        publish own heartbeat on the hb channel
        '''

        self._hb_pub.put('heartbeat', msg={'uid' : self.uid})
      # self._log.debug('hb_cb %s: put %s', self.uid, self.uid)


    # --------------------------------------------------------------------------
    #
    def _hb_term_cb(self, uid=None):

        self._log.debug('hb_term %s: %s died', self.uid, uid)
        self._prof.prof('term', uid=self._uid)

        # FIXME: restart goes here

        # NOTE: returning `False` indicates failure to recover.  The HB will
        #       terminate and suicidally kill the very process it is living in.
        #       Make sure all required cleanup is done at this point!

        return None


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        return self._uid


    # --------------------------------------------------------------------------
    #
    @property
    def cfg(self):
        return self._cfg


    # --------------------------------------------------------------------------
    #
    def start_bridges(self, cfg=None):
        '''
        check if any bridges are defined under `cfg['bridges']` and start them
        '''

        self._prof.prof('start_bridges_start', uid=self._uid)

        timeout = self._cfg.heartbeat.timeout

        if cfg is None:
            cfg = self._cfg

        for bname, bcfg in cfg.get('bridges', {}).items():

            bcfg.uid         = bname
            bcfg.channel     = bname
            bcfg.cmgr        = self.uid
            bcfg.sid         = cfg.sid
            bcfg.path        = cfg.path
            bcfg.heartbeat   = cfg.heartbeat

            fname = '%s/%s.json' % (cfg.path, bcfg.uid)
            bcfg.write(fname)

            self._log.info('create  bridge %s [%s]', bname, bcfg.uid)

            out, err, ret = ru.sh_callout('radical-pilot-bridge %s' % fname)
            self._log.debug('bridge startup out: %s', out)
            self._log.debug('bridge startup err: %s', err)
            if ret:
                raise RuntimeError('bridge startup failed')

            self._uids.append(bcfg.uid)
            self._log.info('created bridge %s [%s]', bname, bcfg.uid)

        # all bridges should start now, for their heartbeats
        # to appear.
      # self._log.debug('wait   for %s', self._uids)
        failed = self._hb.wait_startup(self._uids, timeout=timeout)
      # self._log.debug('waited for %s: %s', self._uids, failed)
        if failed:
            raise RuntimeError('could not start all bridges %s' % failed)

        self._prof.prof('start_bridges_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def start_components(self, cfg=None):
        '''
        check if any components are defined under `cfg['components']`
        and start them
        '''

        self._prof.prof('start_components_start', uid=self._uid)

        timeout = self._cfg.heartbeat.timeout

        if cfg is None:
            cfg = self._cfg

        # we pass a copy of the complete session config to all components, but
        # merge it into the component specific config settings (no overwrite),
        # and then remove the `bridges` and `components` sections
        #
        scfg = ru.Config(cfg=cfg)
        if 'bridges'    in scfg: del(scfg['bridges'])
        if 'components' in scfg: del(scfg['components'])

        for cname, ccfg in cfg.get('components', {}).items():

            for _ in range(ccfg.get('count', 1)):

                ccfg.uid         = ru.generate_id(cname, ns=self._sid)
                ccfg.cmgr        = self.uid
                ccfg.kind        = cname
                ccfg.sid         = cfg.sid
                ccfg.base        = cfg.base
                ccfg.path        = cfg.path
                ccfg.heartbeat   = cfg.heartbeat

                ccfg.merge(scfg, policy=ru.PRESERVE, log=self._log)

                fname = '%s/%s.json' % (cfg.path, ccfg.uid)
                ccfg.write(fname)

                self._log.info('create  component %s [%s]', cname, ccfg.uid)

                out, err, ret = ru.sh_callout('radical-pilot-component %s' % fname)
                self._log.debug('out: %s' , out)
                self._log.debug('err: %s' , err)
                if ret:
                    raise RuntimeError('bridge startup failed')

                self._uids.append(ccfg.uid)
                self._log.info('created component %s [%s]', cname, ccfg.uid)

        # all components should start now, for their heartbeats
        # to appear.
        failed = self._hb.wait_startup(self._uids, timeout=timeout * 10)
        if failed:
            raise RuntimeError('could not start all components %s' % failed)

        self._prof.prof('start_components_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def close(self):

        self._prof.prof('close', uid=self._uid)

        self._hb_bridge.stop()
        self._hb.stop()


# ------------------------------------------------------------------------------
#
class Component(object):
    '''
    This class provides the basic structure for any RP component which operates
    on stateful things.  It provides means to:

      - define input channels on which to receive new things in certain states
      - define work methods which operate on the things to advance their state
      - define output channels to which to send the things after working on them
      - define notification channels over which messages with other components
        can be exchanged (publish/subscriber channels)

    All low level communication is handled by the base class -- deriving classes
    will register the respective channels, valid state transitions, and work
    methods.  When a 'thing' is received, the component is assumed to have full
    ownership over it, and that no other component will change the 'thing's
    state during that time.

    The main event loop of the component -- `work()` -- is executed on `run()`
    and will not terminate on its own, unless it encounters a fatal error.

    Components inheriting this class should attempt not to use shared
    resources.  That will ensure that multiple instances of the component can
    coexist for higher overall system throughput.  Should access to shared
    resources be necessary, it will require some locking mechanism across
    process boundaries.

    This approach should ensure that

      - 'thing's are always in a well defined state;
      - components are simple and focus on the semantics of 'thing' state
        progression;
      - no state races can occur on 'thing' state progression;
      - only valid state transitions can be enacted (given correct declaration
        of the component's semantics);
      - the overall system is performant and scalable.

    Inheriting classes SHOULD overload the following methods:

      - `initialize()`:
        - set up the component state for operation
        - register input/output/notification channels
        - register work methods
        - register callbacks to be invoked on state notification
        - the component will terminate if this method raises an exception.

      - `work()`
        - called in the main loop of the component process, on all entities
          arriving on input channels.  The component will *not* terminate if
          this method raises an exception.  For termination, `terminate()` must
          be called.

      - `finalize()`
        - tear down the component (close threads, unregister resources, etc).

    Inheriting classes MUST call the constructor:

        class StagingComponent(rpu.Component):
            def __init__(self, cfg, session):
                rpu.Component.__init__(self, cfg, session)


    A component thus must be passed a configuration (either as a path pointing
    to a file name to be opened as `ru.Config`, or as a pre-populated
    `ru.Config` instance).  That config MUST contain a session ID (`sid`) for
    the session under which to run this component, and a uid for the component
    itself which MUST be unique within the scope of the given session.  It MUST
    further contain information about the session's heartbeat ZMQ pubsub channel
    (`hb_pub`, `hb_sub`) on which heartbeats are sent and received for lifetime
    management.  All components and the session will continuously sent
    heartbeat messages on that channel - missing heartbeats will by default lead
    to session termination.

    The config MAY contain `bridges` and `component` sections.  If those exist,
    the component will start the communication bridges and the components
    specified therein, and is then considered an owner of those components and
    bridges.  As such, it much watch the HB channel for heartbeats from those
    components, and must terminate itself if those go AWOL.

    Further, the class must implement the registered work methods, with
    a signature of:

        work(self, things)

    The method is expected to change the state of the 'thing's given.  'Thing's
    will not be pushed to outgoing channels automatically -- to do so, the work
    method has to call (see call documentation for other options):

        self.advance(thing)

    Until that method is called, the component is considered the sole owner of
    the 'thing's.  After that method is called, the 'thing's are considered
    disowned by the component.  If, however, components return from the work
    methods without calling advance on the given 'thing's, then the component
    keeps ownership of the 'thing's to advance it asynchronously at a later
    point in time.  That implies that a component can collect ownership over an
    arbitrary number of 'thing's over time, and they can be advanced at the
    component's discretion.

    The component process is a stand-alone daemon process which runs outside of
    Python's multiprocessing domain.  As such, it can freely use Python's
    multithreading (and it extensively does so by default) - but developers
    should be aware that spawning additional *processes* in this component is
    discouraged, as Python's process management is not playing well with it's
    multithreading implementation.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):
        '''
        This constructor MUST be called by inheriting classes, as it specifies
        the operation mode of the component: components can spawn a child
        process, or not.

        If a child will be spawned later, then the child process state can be
        initialized by overloading the`initialize()` method.

        Note that this policy should be strictly followed by all derived
        classes, as we will otherwise carry state over the process fork.  That
        can become nasty if the state included any form of locking (like, for
        profiling or locking).

        The symmetric teardown methods are called `finalize()`.

        Constructors of inheriting components *may* call start() in their
        constructor.
        '''

        # NOTE: a fork will not duplicate any threads of the parent process --
        #       but it will duplicate any locks which are shared between the
        #       parent process and its threads -- and those locks might be in
        #       any state at this point.  As such, each child has to make
        #       sure to never, ever, use any of the inherited locks, but instead
        #       to create it's own set of locks in self.initialize.

        self._cfg     = cfg
        self._uid     = cfg.uid
        self._session = session

        # we always need an UID
        assert(self._uid), 'Component needs a uid (%s)' % type(self)

        # state we carry over the fork
        self._debug      = cfg.get('debug')
        self._owner      = cfg.get('owner', self.uid)
        self._ctype      = "%s.%s" % (self.__class__.__module__,
                                      self.__class__.__name__)
        self._number     = cfg.get('number', 0)
        self._name       = cfg.get('name.%s' %  self._number,
                                   '%s.%s'   % (self._ctype, self._number))

        self._bridges    = list()       # communication bridges
        self._components = list()       # sub-components
        self._inputs     = dict()       # queues to get things from
        self._outputs    = dict()       # queues to send things to
        self._workers    = dict()       # methods to work on things
        self._publishers = dict()       # channels to send notifications to
        self._threads    = dict()       # subscriber and idler threads
        self._cb_lock    = ru.RLock('comp.cb_lock.%s' % self._name)
                                        # guard threaded callback invokations

        self._subscribers = dict()      # ZMQ Subscriber classes

        if self._owner == self.uid:
            self._owner = 'root'

        self._prof = self._session._get_profiler(name=self.uid)
        self._rep  = self._session._get_reporter(name=self.uid)
        self._log  = self._session._get_logger  (name=self.uid,
                                                 level=self._debug)
      # self._prof.register_timing(name='component_lifetime',
      #                            scope='uid=%s' % self.uid,
      #                            start='component_start',
      #                            stop='component_stop')
      # self._prof.register_timing(name='entity_runtime',
      #                            scope='entity',
      #                            start='get',
      #                            stop=['put', 'drop'])
        self._prof.prof('init1', uid=self._uid, msg=self._prof.path)

        self._q    = None
        self._in   = None
        self._out  = None
        self._poll = None
        self._ctx  = None

        self._thread = None
        self._term   = mt.Event()


    # --------------------------------------------------------------------------
    #
    def start(self):

        sync = mt.Event()
        self._thread = mt.Thread(target=self._worker_thread, args=[sync])
        self._thread.daemon = True
        self._thread.start()

        while not sync.is_set():

            if not self._thread.is_alive():
                raise RuntimeError('worker thread died during initialization')

            time.sleep(0.1)

        assert(self._thread.is_alive())


    # --------------------------------------------------------------------------
    #
    def _worker_thread(self, sync):

        try:
            self._initialize()

        except Exception:
            self._log.exception('worker thread initialization failed')
            return

        sync.set()

        while not self._term.is_set():
            try:
                ret = self.work_cb()
                if not ret:
                    break
            except:
                self._log.exception('work cb error [ignored]')


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def create(cfg, session):

        # TODO:  We keep this static typemap for component startup. The map
        #        should really be derived from rp module inspection via an
        #        `ru.PluginManager`.
        #
        from radical.pilot import worker       as rpw
        from radical.pilot import pmgr         as rppm
        from radical.pilot import umgr         as rpum
        from radical.pilot import agent        as rpa
        from radical.pilot import task_overlay as rpt
      # from radical.pilot import constants as rpc

        comp = {
                rpc.WORKER                         : rpt.Worker,
                rpc.UPDATE_WORKER                  : rpw.Update,

                rpc.PMGR_LAUNCHING_COMPONENT       : rppm.Launching,

                rpc.UMGR_STAGING_INPUT_COMPONENT   : rpum.Input,
                rpc.UMGR_SCHEDULING_COMPONENT      : rpum.Scheduler,
                rpc.UMGR_STAGING_OUTPUT_COMPONENT  : rpum.Output,

                rpc.AGENT_STAGING_INPUT_COMPONENT  : rpa.Input,
                rpc.AGENT_SCHEDULING_COMPONENT     : rpa.Scheduler,
                rpc.AGENT_EXECUTING_COMPONENT      : rpa.Executing,
                rpc.AGENT_STAGING_OUTPUT_COMPONENT : rpa.Output

               }

        assert(cfg.kind in comp), '%s not in %s' % (cfg.kind, list(comp.keys()))

        return comp[cfg.kind].create(cfg, session)


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        return "%s <%s> [%s]" % (self.uid, self.__class__.__name__, self._owner)


    # --------------------------------------------------------------------------
    #
    def _cancel_monitor_cb(self, topic, msg):
        '''
        We listen on the control channel for cancel requests, and append any
        found UIDs to our cancel list.
        '''

        # FIXME: We do not check for types of things to cancel - the UIDs are
        #        supposed to be unique.  That abstraction however breaks as we
        #        currently have no abstract 'cancel' command, but instead use
        #        'cancel_units'.

        self._log.debug('command incoming: %s', msg)

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_units':

            uids = arg['uids']

            if not isinstance(uids, list):
                uids = [uids]

            self._log.debug('register for cancellation: %s', uids)

            with self._cancel_lock:
                self._cancel_list += uids

        if cmd == 'terminate':
            self._log.info('got termination command')
            self.stop()

        else:
            self._log.debug('command ignored: %s', cmd)

        return True


    # --------------------------------------------------------------------------
    #
    @property
    def cfg(self):
        return copy.deepcopy(self._cfg)

    @property
    def session(self):
        return self._session

    @property
    def uid(self):
        return self._uid

    @property
    def ctype(self):
        return self._ctype


    # --------------------------------------------------------------------------
    #
    def _initialize(self):
        '''
        initialization of component base class goes here
        '''
        # components can always publish logs, state updates and control messages
     #  self.register_publisher(rpc.LOG_PUBSUB)
        self.register_publisher(rpc.STATE_PUBSUB)
        self.register_publisher(rpc.CONTROL_PUBSUB)

        # set controller callback to handle cancellation requests
        self._cancel_list = list()
        self._cancel_lock = ru.RLock('comp.cancel_lock.%s' % self._uid)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self._cancel_monitor_cb)

        # call component level initialize
        self.initialize()
        self._prof.prof('component_init')


    def initialize(self):
        pass  # can be overloaded


    # --------------------------------------------------------------------------
    #
    def _finalize(self):

        self._log.debug('_finalize()')

        # call component level finalize, before we tear down channels
        self.finalize()

        for thread in self._threads.values():
            thread.stop()

        self._log.debug('%s close prof', self.uid)
        try:
            self._prof.prof('component_final')
            self._prof.close()
        except Exception:
            pass


    def finalize(self):
        pass  # can be overloaded


    # --------------------------------------------------------------------------
    #
    def stop(self, timeout=None):                                         # noqa
        '''
        We need to terminate and join all threads, close all communication
        channels, etc.  But we trust on the correct invocation of the finalizers
        to do all this, and thus here only forward the stop request to the base
        class.
        '''

        #  FIXME: implement timeout, or remove parameter
        #   (pylint W0613 should be removed if changes to timeout are applied)

        self._log.info('stop %s (%s : %s) [%s]', self.uid, os.getpid(),
                       ru.get_thread_name(), ru.get_caller_name())

        self._term.set()
        self._finalize()


    # --------------------------------------------------------------------------
    #
    def register_input(self, states, input, worker=None):
        '''
        Using this method, the component can be connected to a queue on which
        things are received to be worked upon.  The given set of states (which
        can be a single state or a list of states) will trigger an assert check
        upon thing arrival.

        This method will further associate a thing state with a specific worker.
        Upon thing arrival, the thing state will be used to lookup the
        respective worker, and the thing will be handed over.  Workers should
        call self.advance(thing), in order to push the thing toward the next
        component.  If, for some reason, that is not possible before the worker
        returns, the component will retain ownership of the thing, and should
        call advance() asynchronously at a later point in time.

        Worker invocation is synchronous, ie. the main event loop will only
        check for the next thing once the worker method returns.
        '''

        if not isinstance(states, list):
            states = [states]

        name = '%s.%s.%s' % (self.uid, worker.__name__, '_'.join(states))

        if name in self._inputs:
            raise ValueError('input %s already registered' % name)

        # dig the addresses from the bridge's config file
        fname = '%s/%s.cfg' % (self._cfg.path, input)
        cfg   = ru.read_json(fname)

        self._inputs[name] = {'queue'  : ru.zmq.Getter(input, url=cfg['get'],
                                         log=self._log),
                              'states' : states}

        self._log.debug('registered input %s', name)

        # we want exactly one worker associated with a state -- but a worker can
        # be responsible for multiple states
        for state in states:

            self._log.debug('%s register input %s: %s', self.uid, state, name)

            if state in self._workers:
                self._log.warn("%s replaces worker for %s (%s)"
                        % (self.uid, state, self._workers[state]))
            self._workers[state] = worker

            self._log.debug('registered worker %s [%s]', worker.__name__, state)


    # --------------------------------------------------------------------------
    #
    def unregister_input(self, states, input, worker):
        '''
        This methods is the inverse to the 'register_input()' method.
        '''

        if not isinstance(states, list):
            states = [states]

        name = '%s.%s.%s' % (self.uid, worker.__name__, '_'.join(states))

        if name not in self._inputs:
            self._log.warn('input %s not registered', name)
            return

        self._inputs[name]['queue'].stop()
        del(self._inputs[name])
        self._log.debug('unregistered input %s', name)

        for state in states:

            self._log.debug('%s unregister input %s: %s', self.uid, state, name)
            if state not in self._workers:
                self._log.warn('%s input %s unknown', worker.__name__, state)

            del(self._workers[state])
            self._log.debug('unregistered worker %s [%s]', worker.__name__, state)


    # --------------------------------------------------------------------------
    #
    def register_output(self, states, output=None):
        '''
        Using this method, the component can be connected to a queue to which
        things are sent after being worked upon.  The given set of states (which
        can be a single state or a list of states) will trigger an assert check
        upon thing departure.

        If a state but no output is specified, we assume that the state is
        final, and the thing is then considered 'dropped' on calling advance() on
        it.  The advance() will trigger a state notification though, and then
        mark the drop in the log.  No other component should ever again work on
        such a final thing.  It is the responsibility of the component to make
        sure that the thing is in fact in a final state.
        '''

        if not isinstance(states, list):
            states = [states]

        for state in states:

            self._log.debug('%s register output %s:%s', self.uid, state, output)

            # we want a *unique* output queue for each state.
            if state in self._outputs:
                self._log.warn("%s replaces output for %s : %s -> %s"
                        % (self.uid, state, self._outputs[state], output))

            if not output:
                # this indicates a final state
                self._outputs[state] = None

            else:
                # non-final state, ie. we want a queue to push to
                # dig the addresses from the bridge's config file
                fname = '%s/%s.cfg' % (self._cfg.path, output)
                cfg   = ru.read_json(fname)

                self._outputs[state] = ru.zmq.Putter(output, url=cfg['put'])



    # --------------------------------------------------------------------------
    #
    def unregister_output(self, states):
        '''
        this removes any outputs registerd for the given states.
        '''

        if not isinstance(states, list):
            states = [states]

        for state in states:
            self._log.debug('TERM : %s unregister output %s', self.uid, state)

            if state not in self._outputs:
                self._log.warn('state %s has no output registered',  state)
              # raise ValueError('state %s has no output registered' % state)
                continue

            del(self._outputs[state])
            self._log.debug('unregistered output for %s', state)


    # --------------------------------------------------------------------------
    #
    def register_timed_cb(self, cb, cb_data=None, timer=None):
        '''
        Idle callbacks are invoked at regular intervals -- they are guaranteed
        to *not* be called more frequently than 'timer' seconds, no promise is
        made on a minimal call frequency.  The intent for these callbacks is to
        run lightweight work in semi-regular intervals.
        '''

        name = "%s.idler.%s" % (self.uid, cb.__name__)
        self._log.debug('START: %s register idler %s', self.uid, name)

        with self._cb_lock:
            if name in self._threads:
                raise ValueError('cb %s already registered' % cb.__name__)

            if timer is None: timer = 0.0  # NOTE: busy idle loop
            else            : timer = float(timer)

            # create a separate thread per idle cb, and let it be watched by the
            # ru.Process base class
            #
            # ------------------------------------------------------------------
            # NOTE: idle timing is a tricky beast: if we sleep for too long,
            #       then we have to wait that long on stop() for the thread to
            #       get active again and terminate/join.  So we always sleep
            #       just a little, and explicitly check if sufficient time has
            #       passed to activate the callback.
            class Idler(mt.Thread):

                # --------------------------------------------------------------
                def __init__(self, name, log, timer, cb, cb_data, cb_lock):
                    self._name    = name
                    self._log     = log
                    self._timeout = timer
                    self._cb      = cb
                    self._cb_data = cb_data
                    self._cb_lock = cb_lock
                    self._last    = 0.0
                    self._term    = mt.Event()

                    super(Idler, self).__init__()
                    self.daemon = True
                    self.start()

                def stop(self):
                    self._term.set()

                def run(self):
                    try:
                        self._log.debug('start idle thread: %s', self._cb)
                        ret = True
                        while ret and not self._term.is_set():
                            if self._timeout and \
                               self._timeout > (time.time() - self._last):
                                # not yet
                                time.sleep(0.1)  # FIXME: make configurable
                                continue

                            with self._cb_lock:
                                if self._cb_data is not None:
                                    ret = self._cb(cb_data=self._cb_data)
                                else:
                                    ret = self._cb()
                            if self._timeout:
                                self._last = time.time()
                    except:
                        self._log.exception('idle thread failed: %s', self._cb)
            # ------------------------------------------------------------------

            idler = Idler(name=name, timer=timer, log=self._log,
                          cb=cb, cb_data=cb_data, cb_lock=self._cb_lock)
            self._threads[name] = idler

        self._log.debug('%s registered idler %s', self.uid, name)


    # --------------------------------------------------------------------------
    #
    def unregister_timed_cb(self, cb):
        '''
        This method is reverts the register_timed_cb() above: it
        removes an idler from the component, and will terminate the
        respective thread.
        '''

        name = "%s.idler.%s" % (self.uid, cb.__name__)
        self._log.debug('TERM : %s unregister idler %s', self.uid, name)

        with self._cb_lock:

            if name not in self._threads:
                self._log.warn('timed cb %s is not registered', name)
              # raise ValueError('%s is not registered' % name)
                return

            self._threads[name].stop()  # implies join
            del(self._threads[name])

        self._log.debug("TERM : %s unregistered idler %s", self.uid, name)


    # --------------------------------------------------------------------------
    #
    def register_publisher(self, pubsub):
        '''
        Using this method, the component can registered itself to be a publisher
        of notifications on the given pubsub channel.
        '''

        assert(pubsub not in self._publishers)

        # dig the addresses from the bridge's config file
        fname = '%s/%s.cfg' % (self._cfg.path, pubsub)
        cfg   = ru.read_json(fname)

        self._publishers[pubsub] = ru.zmq.Publisher(channel=pubsub,
                                                    url=cfg['pub'],
                                                    log=self._log,
                                                    prof=self._prof)

        self._log.debug('registered publisher for %s', pubsub)


    # --------------------------------------------------------------------------
    #
    def register_subscriber(self, pubsub, cb):
        '''
        This method is complementary to the register_publisher() above: it
        registers a subscription to a pubsub channel.  If a notification
        is received on thag channel, the registered callback will be
        invoked.  The callback MUST have one of the signatures:

          callback(topic, msg)

        where 'topic' is set to the name of the pubsub channel.

        The subscription will be handled in a separate thread, which implies
        that the callback invocation will also happen in that thread.  It is the
        caller's responsibility to ensure thread safety during callback
        invocation.
        '''

        # dig the addresses from the bridge's config file
        fname = '%s/%s.cfg' % (self._cfg.path, pubsub)
        cfg   = ru.read_json(fname)

        if pubsub not in self._subscribers:
            self._subscribers[pubsub] = ru.zmq.Subscriber(channel=pubsub,
                                                          url=cfg['sub'],
                                                          log=self._log,
                                                          prof=self._prof)

        self._subscribers[pubsub].subscribe(topic=pubsub, cb=cb,
                                            lock=self._cb_lock)


    # --------------------------------------------------------------------------
    #
    def work_cb(self):
        '''
        This is the main routine of the component, as it runs in the component
        process.  It will first initialize the component in the process context.
        Then it will attempt to get new things from all input queues
        (round-robin).  For each thing received, it will route that thing to the
        respective worker method.  Once the thing is worked upon, the next
        attempt on getting a thing is up.
        '''

        # if no action occurs in this iteration, idle
        if not self._inputs:
            time.sleep(0.1)
            return True

        for name in self._inputs:
            input  = self._inputs[name]['queue']
            states = self._inputs[name]['states']

            # FIXME: a simple, 1-thing caching mechanism would likely
            #        remove the req/res overhead completely (for any
            #        non-trivial worker).
            things = input.get_nowait(500)  # in microseconds
            things = ru.as_list(things)

            if not things:
                return True

            # the worker target depends on the state of things, so we
            # need to sort the things into buckets by state before
            # pushing them
            buckets = dict()
            for thing in things:
                state = thing['state']
                uid   = thing['uid']
                self._prof.prof('get', uid=uid, state=state)

                if state not in buckets:
                    buckets[state] = list()
                buckets[state].append(thing)

            # We now can push bulks of things to the workers

            for state,things in buckets.items():

                assert(state in states), 'inconsistent state'
                assert(state in self._workers), 'no worker for state %s' % state

                try:
                    to_cancel = list()
                    for thing in things:
                        uid   = thing['uid']
                        ttype = thing['type']
                        state = thing['state']

                        # FIXME: this can become expensive over time
                        #        if the cancel list is never cleaned
                        if uid in self._cancel_list:
                            with self._cancel_lock:
                                self._cancel_list.remove(uid)
                            to_cancel.append(thing)

                        self._log.debug('got %s (%s)', ttype, uid)

                    if to_cancel:
                        self.advance(to_cancel, rps.CANCELED, publish=True,
                                                              push=False)
                    with self._cb_lock:
                        self._workers[state](things)

                except Exception:

                    # this is not fatal -- only the 'things' fail, not
                    # the component
                    self._log.exception("work %s failed", self._workers[state])
                    self.advance(things, rps.FAILED, publish=True, push=False)


        # keep work_cb registered
        return True


    # --------------------------------------------------------------------------
    #
    def advance(self, things, state=None, publish=True, push=False, ts=None,
                      prof=True):
        '''
        Things which have been operated upon are pushed down into the queues
        again, only to be picked up by the next component, according to their
        state model.  This method will update the thing state, and push it into
        the output queue registered as target for that state.

        things:  list of things to advance
        state:   new state to set for the things
        publish: determine if state update notifications should be issued
        push:    determine if things should be pushed to outputs
        prof:    determine if state advance creates a profile event
                 (publish, and push are always profiled)

        'Things' are expected to be a dictionary, and to have 'state', 'uid' and
        optionally 'type' set.

        If 'thing' contains an '$all' key, the complete dict is published;
        otherwise, *only the state* is published.

        This is evaluated in self.publish.
        '''

        if not ts:
            ts = time.time()

        things = ru.as_list(things)

        if not things:
            return

        self._log.debug('advance bulk: %s [%s, %s]', len(things), push, publish)

        # assign state, sort things by state
        buckets = dict()
        for thing in things:

            uid = thing['uid']

          # if thing['type'] not in ['unit', 'pilot']:
          #     raise TypeError("thing has unknown type (%s)" % uid)

            if state:
                # state advance done here
                thing['state'] = state

            _state = thing['state']

            if prof:
                self._prof.prof('advance', uid=uid, state=_state, ts=ts)

            if _state not in buckets:
                buckets[_state] = list()
            buckets[_state].append(thing)

        # should we publish state information on the state pubsub?
        if publish:

            to_publish = list()

            # If '$all' is set, we update the complete thing_dict.
            # Things in final state are also published in full.
            # If '$set' is set, we also publish all keys listed in there.
            # In all other cases, we only send 'uid', 'type' and 'state'.
            for thing in things:
                if '$all' in thing:
                    del(thing['$all'])
                    to_publish.append(thing)

                elif thing['state'] in rps.FINAL:
                    to_publish.append(thing)

                else:
                    tmp = {'uid'   : thing['uid'],
                           'type'  : thing['type'],
                           'state' : thing['state']}
                    for key in thing.get('$set', []):
                        tmp[key] = thing[key]
                    to_publish.append(tmp)

            self.publish(rpc.STATE_PUBSUB, {'cmd': 'update', 'arg': to_publish})

          # ts = time.time()
          # for thing in things:
          #     self._prof.prof('publish', uid=thing['uid'],
          #                     state=thing['state'], ts=ts)

        # never carry $all across component boundaries!
        for thing in things:
            if '$all' in thing:
                del(thing['$all'])

        # should we push things downstream, to the next component
        if push:

            # the push target depends on the state of things, so we need to sort
            # the things into buckets by state before pushing them
            # now we can push the buckets as bulks
            for _state,_things in buckets.items():

              # ts = time.time()
                if _state in rps.FINAL:
                    # things in final state are dropped
                    for thing in _things:
                        self._log.debug('final %s [%s]', thing['uid'], _state)
                        self._prof.prof('drop', uid=thing['uid'], state=_state,
                                        ts=ts)
                    continue

                if _state not in self._outputs:
                    # unknown target state -- error
                    for thing in _things:
                        import pprint
                        self._log.debug('%s', pprint.pformat(self._outputs))
                        self._log.debug("lost  %s [%s]", thing['uid'], _state)
                        self._prof.prof('lost', uid=thing['uid'], state=_state,
                                        ts=ts)
                    continue

                if not self._outputs[_state]:
                    # empty output -- drop thing
                    for thing in _things:
                        self._log.debug('drop  %s [%s]', thing['uid'], _state)
                        self._prof.prof('drop', uid=thing['uid'], state=_state,
                                        ts=ts)
                    continue

                output = self._outputs[_state]

                # push the thing down the drain
                self._log.debug('put bulk %s: %s', _state, len(_things))
                output.put(_things)

                ts = time.time()
                for thing in _things:
                    self._prof.prof('put', uid=thing['uid'], state=_state,
                                    msg=output.name, ts=ts)


    # --------------------------------------------------------------------------
    #
    def publish(self, pubsub, msg):
        '''
        push information into a publication channel
        '''

        if pubsub not in self._publishers:
            self._log.warn("can't route msg for '%s': %s" % (pubsub,
                                                 list(self._publishers.keys())))
            return

          # raise RuntimeError("can't route '%s' notification: %s" % (pubsub,
          #     self._publishers.keys()))

        if not self._publishers[pubsub]:
            raise RuntimeError("no msg route for '%s': %s" % (pubsub, msg))

        self._publishers[pubsub].put(pubsub, msg)


# ------------------------------------------------------------------------------
#
class Worker(Component):
    '''
    A Worker is a Component which cannot change the state of the thing it
    handles.  Workers are employed as helper classes to mediate between
    components, between components and database, and between components and
    notification channels.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        Component.__init__(self, cfg=cfg, session=session)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2017, http://radical.rutgers.edu"
__license__   = "MIT"


import copy

import radical.utils as ru

from .base import AgentSchedulingComponent


# ------------------------------------------------------------------------------
#
class Hombre(AgentSchedulingComponent):
    '''
    HOMBRE: HOMogeneous Bag-of-task REsource allocator.  Don't kill me...

    The scheduler implementation manages a set of resources (cores).  When it
    gets the first scheduling request, it will assume that to be representative
    *for all future sccheduling requests*, ie. it assumes that all tasks are
    homogeneous.  It thus splits the resources into equal sized chunks and
    assigns them to the incoming requests.  The chunking is fast, no decisions
    need to be made, the scheduler really just hands out pre-defined chunks,
    which is relatively quick - that's the point, to trade generality with
    performance.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self.nodes = None

        AgentSchedulingComponent.__init__(self, cfg, session)

        # homogeneous workloads are always uniform
        self._uniform_wl = True


    # --------------------------------------------------------------------------
    #
    # FIXME: this should not be overloaded here, but in the base class
    #
    def finalize_child(self):

        # make sure that parent finalizers are called
        super(Hombre, self).finalize_child()


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # * oversubscribe:
        #   Cray's aprun for example does not allow us to oversubscribe CPU
        #   cores on a node, so we can't, say, run n CPU processes on an n-core
        #   node, and than add one additional process for a GPU application.  If
        #   `oversubscribe` is set to False (which is the default for now),
        #   we'll prevent that behavior by allocating one additional CPU core
        #   for each set of requested GPU processes.
        self._oversubscribe = self._cfg.get('oversubscribe', True)

        if not self._oversubscribe:
            raise ValueError('HOMBRE needs oversubscription enabled')

        # NOTE: We delay the actual configuration until we received the first
        #       unit to schedule - at that point we can slice and dice the
        #       resources into suitable static slots.
        self._configured = False

        self.free = list()     # declare for early debug output


    # --------------------------------------------------------------------------
    #
    def _delayed_configure(self, cud):

        if self._configured:
            return

        self.chunk  = {'cpu_processes'    : cud['cpu_processes'   ],
                       'cpu_process_type' : cud['cpu_process_type'],
                       'cpu_threads'      : cud['cpu_threads'     ],
                       'cpu_thread_type'  : cud['cpu_thread_type' ],

                       'gpu_processes'    : cud['gpu_processes'   ],
                       'gpu_process_type' : cud['gpu_process_type'],
                       'gpu_threads'      : cud['gpu_threads'     ],
                       'gpu_thread_type'  : cud['gpu_thread_type' ],
                       }

        self.cpn     = self._rm_cores_per_node
        self.gpn     = self._rm_gpus_per_node

        self.free    = list()     # list of free chunks
        self.lock    = ru.Lock()  # lock for the above list

        cores_needed = cud['cpu_processes'] * cud['cpu_threads']
        gpus_needed  = cud['gpu_processes']

        # check if we need single or multi-node chunks
        if  cud['cpu_process_type'] != 'MPI' and \
            cud['gpu_process_type'] != 'MPI' :

            # single node task - check if it fits
            if cores_needed > self.cpn or \
               gpus_needed  > self.gpn:
                raise ValueError('unit does not fit on node')


        # ---------------------------------------------------------------------
        # create as many equal sized chunks from the available nodes as
        # possible, and put them into the `free` list.  The actual scheduling
        # algorithm will blindly pick chunks from that list whenever a new CUD
        # arrives.
        cblock   = cud['cpu_threads']
        ncblocks = cud['cpu_processes']
        cblocks  = list()
        cidx     = 0

        while cidx + cblock <= self.cpn:
            cblocks.append(list(range(cidx,cidx + cblock)))
            cidx += cblock

        gblock   = 1
        ngblocks = cud['gpu_processes']
        gblocks  = list()
        gidx     = 0
        while gidx + gblock <= self.gpn:
            gblocks.append(list(range(gidx,gidx + gblock)))
            gidx += gblock

        self._log.debug('core blocks %s', cblocks)
        self._log.debug('gpu  blocks %s', gblocks)

        for node in self.nodes:
            node['cblocks'] = copy.deepcopy(cblocks)
            node['gblocks'] = copy.deepcopy(gblocks)

        # ----------------------------------------------------------------------
        def next_slot(slot=None):
            if slot:
                del(slot['ncblocks'])
                del(slot['ngblocks'])
                self.free.append(slot)
            return {'nodes'         : list(),
                    'cores_per_node': self.cpn,
                    'gpus_per_node' : self.gpn,
                    'lm_info'       : self._rm_lm_info,
                    'ncblocks'      : 0,
                    'ngblocks'      : 0}
        # ---------------------------------------------------------------------
        nidx   = 0
        nnodes = len(self.nodes)
        slot   = next_slot()
        while nidx < nnodes:


            if  slot['ncblocks'] == ncblocks and \
                slot['ngblocks'] == ngblocks :
                slot = next_slot(slot)

            node  = self.nodes[nidx]
            nuid  = node['uid']
            nname = node['name']
            ok    = True

            while slot['ncblocks'] < ncblocks:
                if node['cblocks']:
                    cblock = node['cblocks'].pop(0)
                    slot['nodes'].append({'name'    : nname,
                                          'uid'     : nuid,
                                          'core_map': [cblock],
                                          'gpu_map' : []})
                    slot['ncblocks'] += 1
                else:
                    ok = False
                    break

            while slot['ngblocks'] < ngblocks:
                if node['gblocks']:

                    # move the process onto core `0` (oversubscribed)
                    # enabled)
                    gblock = node['gblocks'].pop(0)
                    slot['nodes'].append({'name'    : nname,
                                          'uid'     : nuid,
                                          'core_map': [[0]],
                                          'gpu_map' : [gblock]})
                    slot['ngblocks'] += 1
                else:
                    ok = False
                    break

            if ok:
                self.free.append(slot)
                slot = next_slot()
                continue

            nidx += 1


        if  slot['ncblocks'] == ncblocks and \
            slot['ngblocks'] == ngblocks :
            self.free.append(slot)

        if not self.free:
            raise RuntimeError('configuration cannot be used for this workload')

        # run this method only once
        self._configured = True


    # --------------------------------------------------------------------------
    #
    def _allocate_slot(self, cud):
        '''
        This is the main method of this implementation, and is triggered when
        a unit needs to be mapped to a set of cores / gpus.
        '''

      # self._log.debug('=> allocate [%d]', len(self.free))
        self._delayed_configure(cud)

        # ensure that all CUDs require the same amount of reources
        for k,v in list(self.chunk.items()):
            if cud[k] != v:
                raise ValueError('hetbre?  %d != %d' % (v, cud[k]))

      # self._log.debug('find new slot')
        slots = self._find_slots(cud)
        if slots:
            self._log.debug('allocate slot %s', slots['nodes'])
        else:
            self._log.debug('allocate slot %s', slots)

      # self._log.debug('<= allocate [%d]', len(self.free))


        return slots


    # --------------------------------------------------------------------------
    #
    def _release_slot(self, slots):
        '''
        This method is called when previously aquired resources are not needed
        anymore.  `slots` are the resource slots as previously returned by
        `_allocate_slots()`.
        '''

      # self._log.debug('=> release  [%d]', len(self.free))
        self._log.debug('release  slot %s', slots['nodes'])
        with self.lock:
            self.free.append(slots)
      # self._log.debug('<= release  [%d]', len(self.free))


    # --------------------------------------------------------------------------
    #
    def _find_slots(self, cores_requested):

        # check if we have free chunks laying around - return one
        with self.lock:
            if self.free:
                return self.free.pop()
            else:
                return None


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys

import radical.pilot as rp
import radical.utils as ru


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    if   len(sys.argv)  > 2:
        report.exit('Usage:\t%s [resource]\n\n' % sys.argv[0])
    elif len(sys.argv) == 2:
        resource = sys.argv[1]
    else:
        resource = 'local.localhost'

    session = rp.Session()
    try:

        report.info('read config')
        pwd    = os.path.dirname(os.path.abspath(__file__))
        config = ru.read_json('%s/../config.json' % pwd)
        report.ok('>>ok\n')

        report.header('submit pilots')
        pd_init = {'resource'      : resource,
                   'runtime'       : 15,  # pilot runtime (min)
                   'exit_on_error' : True,
                   'project'       :     config[resource]['project'],
                   'queue'         :     config[resource]['queue'],
                   'access_schema' :     config[resource]['schema'],
                   'cores'         : 8
                  }
        pdesc = rp.ComputePilotDescription(pd_init)
        pmgr  = rp.PilotManager(session=session)
        pilot = pmgr.submit_pilots(pdesc)

        report.header('stage data')
        pilot.stage_in({'source': 'client:///gromacs/',
                        'target': 'pilot:///',
                        'action': rp.TRANSFER})
        report.ok('>>ok\n')

        report.header('submit units')

        umgr = rp.UnitManager(session=session)
        umgr.add_pilots(pilot)

        args = 'mdrun -o traj.trr -e ener.edr -s topol.tpr -g mdlog.log ' \
             + '-c outgro -cpo state.cpt'

        tags = {'app-stats'  : 'this_app',
                'constraint' : 'p * t <= 32'}

        cudis = list()
        for f in ['grompp.mdp', 'mdout.mdp', 'start.gro',
                  'topol.top',  'topol.tpr']:
            cudis.append({'source': 'pilot:///gromacs/%s' % f, 
                          'target': 'unit:///%s' % f,
                          'action': rp.LINK})

        n = 2
        n = 2 * 1024  # number of units to run
        report.info('create %d unit description(s)\n\t' % n)
        cuds = list()
        for i in range(0, n):

            # create a new CU description, and fill it.
            # Here we don't use dict initialization.
            cud = rp.ComputeUnitDescription()
            cud.executable       = 'date'
          # cud.arguments        = '1'.split()
          # cud.executable       = 'gmx'
          # cud.arguments        = args.split()
            cud.tags             = tags
            cud.gpu_processes    = 0
            cud.cpu_processes    = 2  # '1-32'
            cud.cpu_threads      = 2  # '1-16'
            cud.cpu_process_type = rp.MPI
            cud.cpu_thread_type  = rp.OpenMP
      #     cud.input_staging    = cudis

            cuds.append(cud)
            report.progress()
        report.ok('>>ok\n')

        umgr.submit_units(cuds)
        report.header('gather results')
        umgr.wait_units()


    finally:
        report.header('finalize')
        session.close(download=True)

    report.header()


# ------------------------------------------------------------------------------

